#!/usr/bin/python
# Copyright (c) 2009 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors: Greg Nichols
#          YangKun <ykun@redhat.com>
#

import os
import sys
import shutil
import time
from xml.dom.minidom import parseString

from rhcert.tags import Constants, DeviceClass, Tags, TestTag, RHEL, Attributes
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput
from rhcert.environment import Environment
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.redhatrelease import RedHatRelease
from rhcert.model.image import Image

from hardwareTest import HardwareTest


class FvTest(HardwareTest):

    def __init__(self, path):
        path = "fvtest/" + path
        HardwareTest.__init__(self, path)

        self.deviceClass = DeviceClass.system
        self.hardwarePlatform = Command("uname -i").getString()
        self.guestName = "hwcert-%s" % self.hardwarePlatform
        self.guestImageFile = 'hwcert-' + self.hardwarePlatform + '.img'
        self.guestConfigFile = 'hwcert-' + self.hardwarePlatform + ".xml"
        self.loopBackDevice = None
        self.dataImageFile = 'hwcertData.img'
        # backend command expected in guest
        self.backendInGuest = "rhcert-cli"
        self.rhcertMajMinVersion = "5.12"
        self.oldrhcertMajMinVersion = "4.5"

        # defaults
        self.hwcertCommandInGuest = self.backendInGuest + " run"
        # subclasses use either testToRun OR tagToRun to specify tests
        self.testToRun = ""
        self.tagToRun = ""
        self.environment = Environment()
        self.guestConfigPath = os.path.join(self.environment.getKVMGuestConfigDirectory(), self.guestConfigFile)
        self.guestConfigDirectory = self.environment.getKVMGuestConfigDirectory()
        self.guestImageDirectory = self.environment.getKVMGuestImageDirectory()
        self.waitTime = self.environment.getFVGuestTimelimit()
        self.proxyURLs = self.environment.getProxyURLs()
        self.redHatRelease = RedHatRelease()
        self.mountPoint = self.environment.getFvImageMountPoint()
        self.enableTargetDevice = False
        self.targetDevice = None
        try:
            Command("mkdir -p " + self.mountPoint).echo()
        except HwCertCommandException, e:
            print "Error: can't create guest image mountpoint " + self.mountPoint
            raise

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.virtualization]

    def configureHypervisor(self):
        """ configure the test for the kvm hypervisor """
        # RHEL 6.0 onwards kvm is supported and Xen is dropped
        self.hypervisor = Constants.kvm 

        # check cpu flags, just to log the results.
        self.__hasVirtualizationCPUFlags(verbose=True)

        print "Testing kvm virtualization"
        self.hypervisor = Constants.kvm
        self.guestConfigFile = 'hwcert-' + self.hardwarePlatform + ".xml"
        self.guestConfigPath = os.path.join(self.environment.getKVMGuestConfigDirectory(), self.guestConfigFile)
        self.guestConfigDirectory = self.environment.getKVMGuestConfigDirectory()
        self.guestImageDirectory = self.environment.getKVMGuestImageDirectory()

        # check and start libvirtd
        try:
            Command("service libvirtd start").echo()
            print "libvirtd is running"
        except HwCertCommandException, e:
            print "Warning: libvirtd may not be running"
            print e

        # check the kvm modules
        try:
            kvmModulesScript = "/etc/sysconfig/modules/kvm.modules"
            if os.path.exists(kvmModulesScript):
                Command(kvmModulesScript).echo()
        except HwCertCommandException, e:
            print "Warning: kvm modules may not be running"
            print e

    def __isHVMCapableViaLibvirt(self):
        """ determine if system is FV/hvm capable, via virsh from libvirt """

        start_libvirtd = Command("service libvirtd start")
        try:
            start_libvirtd.run()
        except HwCertCommandException:
            print "Warning: libvirtd is not running or not installed"
            return False

        virsh = Command("virsh capabilities")
        try:
            virsh.run()
            parser = parseString("".join(virsh.output))
            elements = parser.getElementsByTagName("guest")
            for element in elements:
                osType = element.getElementsByTagName(Tags.os_type)[0].childNodes[0]
                if osType.data == "hvm":
                    return True
        except Exception, e:
            if self.getDebug() != Constants.off:
                print "Note: could not parse virsh capabilities"
                print e

        #otherwise
        return False

    def __isHVMInCapabilitiesFile(self):
        """ try /sys/../capabilities, looking for xen FV support"""
        try:
            capabilities = '/sys/hypervisor/properties/capabilities'
            capabilityFile = open(capabilities)
            self.hypervisorCapabilities = capabilityFile.readline()
            capabilityFile.close()
            if  Constants.hvm in self.hypervisorCapabilities:
                return True
        except IOError:
            if self.getDebug() != Constants.off:
                print "Note: could not read " + capabilities

        #otherwise
        return False

    def __hasVirtualizationCPUFlags(self, verbose=False):
        """ check cpuinfo for virtualization flags """
        cpuflags = Command("fgrep flags /proc/cpuinfo")
        try:
            flags = cpuflags.getString(regex="(vmx|smx|svm)", singleLine=False)
            if verbose or self.getDebug() != Constants.off:
                print "/proc/cpuinfo has %s flag" % flags
            return True
        except HwCertCommandException:
            if verbose or self.getDebug() != Constants.off:
                print "Note: /proc/cpuinfo does not have vmx, smx, or svm flags"

        return False

    def plan(self, planner):
        tests = list()
        osSupport = False
        # if Red Hat and not RT kernel
        if ("Red Hat" in self.redHatRelease.getProduct() 
            and self.redHatRelease.getKernelFlavor() != Constants.rt 
            and self.redHatRelease.getArch() == Constants.x86_64):
                osSupport = True
        #  Fedora x86_64
        elif "Fedora" in self.redHatRelease.getProduct() and self.redHatRelease.getArch() == Constants.x86_64:
            osSupport = True

        # check libvirt, capabilities file, or cpuflags
        if osSupport and (self.__isHVMCapableViaLibvirt() or self.__isHVMInCapabilitiesFile() or self.__hasVirtualizationCPUFlags()):
                test = self.makeCopy()
                tests.append(test)
                # Download the guest XML config file if fv_test is supported
                self.verifyFile(self.guestConfigDirectory, self.guestConfigFile)
        return tests

    def getRequiredRPMs(self):
        rpms = ["wget", "libvirt", "libvirt-python", "qemu-kvm", "qemu-kvm-tools"]
        return rpms

    def verifyGuest(self):
        guestName = None
        try:
            guestName = Command("virsh list").getString(regex=self.guestName, singleLine=False)
            print "Warning: Guest %s is already running" % guestName
            print "Attempting shutdown:"
            Command("virsh destroy %s" % self.guestName).echo()
            guestName = Command("virsh list").getString(regex=self.guestName, singleLine=False)
            print "Error: could not shutdown guest"
            return False
        except:
            print "Verified that guest " + self.guestName + " is not running"
            pass

        return self.verifyGuestFiles()


    def verifyGuestFiles(self):
        if not self.verifyFile(self.guestImageDirectory, self.dataImageFile):
            return False
        if not self.verifyFile(self.guestImageDirectory, self.guestImageFile):
            return False
        if not self.verifyFile(self.guestConfigDirectory, self.guestConfigFile):
            return False

        print "Guest files verified"
        sys.stdout.flush()
        print "Restarting libvirtd:"
        try:
            Command("service libvirtd restart").echo()
        except HwCertCommandException, exception:
            print "Error: could not restart libvirtd:"
            print exception
            return False

        return True

    def verifyFile(self, destinationDirectory, filename):

        filePath = os.path.join(destinationDirectory, filename)
        if os.path.exists(filePath):
            print "Verified: " + filePath
            return True

        # otherwise
        print "Note: " + filePath + " not found, attempting download"
        return self.installFile(destinationDirectory, filename)

    def installFile(self, destinationDirectory, filename):
        sys.stdout.flush()
        urlsToTry = list()
        tarredFilename = filename + ".tar.bz2"
        if self.getTestServer() and self.getTestServer() != "unknown":
            testServerBaseUrl ="http://%s" % os.path.join(self.getTestServer(), Constants.store, Constants.transfer, Constants.fv_images, RHEL.major)
            testServerNewBaseUrl = "http://%s" % os.path.join(self.getTestServer(), Constants.rhcert_transfer, Constants.fv_images, RHEL.major)
            urlsToTry.append(os.path.join(testServerBaseUrl, self.rhcertMajMinVersion, tarredFilename))
            urlsToTry.append(os.path.join(testServerNewBaseUrl, self.rhcertMajMinVersion, tarredFilename))
            urlsToTry.append(os.path.join(testServerBaseUrl, tarredFilename))
            urlsToTry.append(os.path.join(testServerNewBaseUrl, tarredFilename))
        else:
            print "Warning: no hwcert test server set - can not look for guest files there"
        partnerServerBaseUrl = os.path.join(self.environment.getPartnerServerURL(), Constants.fv_images, RHEL.major)
        urlsToTry.append(os.path.join(partnerServerBaseUrl, self.rhcertMajMinVersion, tarredFilename))
        urlsToTry.append(os.path.join(partnerServerBaseUrl, tarredFilename))

        wget_cmd = "wget "
        if self.proxyURLs:
            # In proxy environment, use proxy of ftp protocol if that is set,
            # else show a warning message and use proxy of http or https protocol.
            # If none of the above protocols found, FAIL the test.
            supported_protocols = ["ftp", "http", "https"]
            proxy = None
            protocol_used = None
            for protocol in supported_protocols:
                if protocol in list(self.proxyURLs.keys()):
                    protocol_used = protocol
                    proxy = self.proxyURLs[protocol]
                    break
            if protocol_used is None:
                print("Error: Proxy settings are missing for ftp protocol, no other valid proxies found. "
                      "Please check settings in /etc/rhcert.xml.")
                print("Error: Failed to download the guest file %s" % tarredFilename)
                return False
            elif protocol_used != "ftp":
                print("Warning: Proxy settings are missing for ftp protocol. Please check settings in /etc/rhcert.xml. "
                      "Using proxy of '%s' protocol" % protocol_used)
            wget_cmd = "wget -e ftp_proxy='%s' " % proxy

        for url in urlsToTry:
            download_cmd = wget_cmd + url
            print("\nTrying to download the guest file using command: %s" % download_cmd)
            if os.system(download_cmd) == 0:
                try:
                    if os.path.exists(tarredFilename):
                        unpackedFilename = Command("tar xvf " + tarredFilename).getString()
                        os.remove(tarredFilename)
                        # move it to the right place
                        shutil.move(unpackedFilename, os.path.join(destinationDirectory, filename))
                        print "Moved: " + unpackedFilename + " to " + os.path.join(destinationDirectory, filename)
                        restorecon = Command("/sbin/restorecon -v " + os.path.join(destinationDirectory, filename))
                        restorecon.echo()
                        print "Downloaded " + filename + " to " + destinationDirectory
                        return True
                    else:
                        print "Note: could not download " + url
                except Exception, e:
                    print "Warning: download " + url + " failed:"
                    print e
            else:
                print "Warning: could not download " + url 

        # otherwise
        print "Error: could not find or download guest file " + tarredFilename
        return False

    def mountDataImage(self):
        try:
            self.loopBackDevice = Command("losetup -f").getString(regex="^(?P<device>/dev/loop[0-9]+)", regexGroup="device", singleLine=False)
            print "Using loopback device " + self.loopBackDevice + " for guest data image"
            loop = os.path.basename(self.loopBackDevice)
            Command("losetup " + self.loopBackDevice + " " + os.path.join(self.guestImageDirectory, self.dataImageFile)).echo()
            try:
                Command("kpartx -av " + self.loopBackDevice).echo()
            except HwCertCommandErrorOutput, exception:
                print "Warning: %s" % exception
                exception.command.printErrors()
            # we may need to wait here - see BZ 874715
            waitTime = 5 # sec
            retries = 6
            while True:
                try:
                    time.sleep(waitTime)
                    partition = Command("ls /dev/mapper").getString(regex="%sp[0-9]" % loop, singleLine=False)
                    break
                except HwCertCommandException, e:
                    if not retries:
                        print "Error: could not create loop device %s" % loop
                        return False
                    # otherwise
                    print "Warning: %s is not ready." % loop
                    retries = retries - 1

            Command("mount /dev/mapper/" + partition + " " + self.mountPoint).run()
            print "Mounted /dev/mapper/" + partition + " on " + self.mountPoint

            # check image meta-data
            self.imageMetaData = None
            try:
                self.imageMetaData = Image(self.mountPoint + "/hwcert-image.xml")
                print self.imageMetaData
            except:
                print "Warning: could not find hwcert image meta-data"

            if self.getDebug()  != Constants.off:
                Command("df -h " + self.mountPoint).echo()
                Command("ls -al " + self.mountPoint).echo()
        except HwCertCommandException, e:
            print "Error: could not mount data image"
            print e
            e.command.printErrors()
            return False
        return True

    def unmountDataImage(self, ignoreErrors=False):
        print "Unmounting " + self.mountPoint
        sys.stdout.flush()
        try:
            Command("umount " + self.mountPoint).echo()
            if self.loopBackDevice:
                print "Deleting loopback device " + self.loopBackDevice + " configuration."
                Command("kpartx -dv " + self.loopBackDevice).echo()
                Command("losetup -d " + self.loopBackDevice).echo()
        except HwCertCommandException, e:
            if not ignoreErrors:
                print "Error: could not unmount data image."
                print e
                e.command.printErrors()
            return False
        return True

    def submitTasks(self):
        # mount the data image and then write test info
        try:
            dataImagePath = os.path.join(self.guestImageDirectory, self.dataImageFile)
            self.unmountDataImage(ignoreErrors=True)
            if not self.mountDataImage():
                return False
            guestScriptFile = self.mountPoint + "/hwcert-guest-script"
            guestScript = open(guestScriptFile, "w")
            guestScript.write("#!/bin/sh\n")
            guestScript.write("# hwcert FV guest script - auto-generated\n")
            # try to install hwcert
            # note: guests expect the data image disk to me /mnt, not /mnt/hwcert
            # see hwcert-client/tools/fv-guest/fv-domU-rc.local
            guestMountPoint = "/mnt"
            guestRPMsPoint = guestMountPoint + "/" + self.rhcertMajMinVersion
            guestOldRPMsPoint = guestMountPoint + "/" + self.oldrhcertMajMinVersion

            # check if directory has already been copied to /root
            rpmDir = "/root/rhcertRPMs"

            # The old images of hwcertData.img will not have the latest rhcertMajMinVersion
            # Copy older RPMs from guestOldRPMsPoint for old images
            guestScript.write("if [ ! -d %s ]; then\n" % rpmDir)
            guestScript.write("    mkdir %s\n" % rpmDir)
            guestScript.write("    if [ -d %s ]; then\n" % guestRPMsPoint)
            guestScript.write("        cp -r %s/* %s\n" %(guestRPMsPoint, rpmDir))
            guestScript.write("    else\n")
            guestScript.write("        cp -r %s/* %s\n" %(guestOldRPMsPoint, rpmDir))
            guestScript.write("    fi\n")
            guestScript.write("fi\n")

            # remove hwcert and rhcert rpms in case they were still there
            for package in ["redhat-certification-hardware", "redhat-certification", "redhat-certification-backend"]:
                guestScript.write("rpm -e %s\n" % package)

            dependencies = ['python-chardet', 'python-ordereddict', 'python-six', 'python-backports', 'python-backports-ssl_match_hostname', 'python-urllib3', 'python-requests', 'mod_wsgi', 'lsscsi']
            for dependency in dependencies:
                guestScript.write("rpm -ivh --nodeps " + rpmDir + "/dependencies/" + dependency + "-*.rpm\n")
            guestScript.write("rpm -ivh --nodeps " + rpmDir + "/redhat-certification-backend*.noarch.rpm\n")
            guestScript.write("rpm -ivh --nodeps " + rpmDir + "/redhat-certification-hardware*.noarch.rpm\n")

            # plan the tests 
            guestScript.write("%s plan --mode auto\n" % self.backendInGuest)

            # run the tests
            options = " --mode auto "
            if self.getTestServer():
                options += ' --server %s ' % self.getTestServer()

            if self.enableTargetDevice and self.targetDevice:
                options += " --device %s " % self.targetDevice

            if self.testToRun:
                options += " --test %s " % self.testToRun
                # for device-specific tests, add them to the plan.
                if self.enableTargetDevice and self.targetDevice:
                    guestScript.write("%s plan --add %s\n" % (self.backendInGuest, options))
            elif self.tagToRun:
                options += " --tag %s " % self.tagToRun

            commandToRun = self.hwcertCommandInGuest + options
            guestScript.write(commandToRun + "\n")

            # copy the results to hwcertData.img
            guestScript.write("mv -f /var/rhcert/results.xml " + guestMountPoint + "/.\n")
            # remove the tmp files
            guestScript.write("echo -e \'y\' | %s clean --mode auto\n" % self.backendInGuest)
            # remove hwcert and rhcert rpms
            for package in ["redhat-certification-hardware", "redhat-certification", "redhat-certification-backend"]:
                guestScript.write("rpm -e %s\n" % package)

            # poweroff
            guestScript.write("poweroff\n")
            guestScript.close()
            os.chmod(guestScriptFile, 0744)
            # older images may call the old script from rc.local, so alias it here
            # note that the data image must have hwcert-client and not v7.
            try:
                os.symlink("./hwcert-guest-script", self.mountPoint + "/v7-guest-script")
            except OSError:
                # link exists, most likely
                pass

            self.unmountDataImage()
            print "Submitted tests: " + commandToRun
            sys.stdout.flush()
            return True
        except HwCertCommandException, e:
            print "Error: could not set guest tests to run"
            print e
            return False




    def runGuest(self):
            
        try:
            Command("virsh start %s" % self.guestName).echo()
            print "FV Guest started ..."
        except HwCertCommandErrorOutput, exception:
            print "Warning: %s" % exception
            exception.command.printErrors()
        except HwCertCommandException, e:
            print "Error: could not start the FV Guest."
            print e
            return False
        print "%s test is running, it takes some time(less than %s minutes), please be patient ..." % (self.getName(), self.waitTime)
        print "You may use \"virsh console %s\" to monitor the testing progress" % self.guestName
        sys.stdout.flush()

        # try and log the guest console if debug is set
        if self.getDebug() != Constants.off:
            try:
                tty = Command("virsh ttyconsole %s" % self.guestName).getString(regex="^/dev.*$", singleLine=False)
                Command("cat %s > guest_console.log &" % tty).echoIgnoreErrors()
            except HwCertCommandException, e:
                print "Warning: debug could not display virsh console log."
                print e
                e.command.printErrors()
        sys.stdout.flush()
        # check the guest's status in every minute, this test should not consume more than 1 hour
        count = 0
        success = True
        while 1:
            if count == self.waitTime:
                try:
                    print "time out: destroying the guest."
                    Command("virsh destroy %s" % self.guestName).echo()
                except HwCertCommandException, exception:
                    print "Warning: %s" % exception
                success = False
                break
            else:
                time.sleep(60)
                try:
                    guestIsRunning = Command("virsh list").getString(regex=self.guestName, singleLine=False)
                    count = count + 1
                except:
                    print "Guest has shutdown"
                    break

        if self.getDebug() != Constants.off:
            try:
                log = open("guest_console.log")
                print "****** guest console log - start ******"
                for line in log.readlines():
                    print "%s" % line,
                print "****** guest console log - end ******"
            except IOError, e:
                print e

        return success

    def getGuestResults(self):
        # mount the data image and see if the result rpm is there
        result = True
        try:
            self.mountDataImage()
            copyCommand = Command("cp %s/%s %s" % (self.mountPoint, self.environment.getResultsFileName(), self.getOutputDirectory()))
            copyCommand.echo()
            print copyCommand.command
            # merge the results
            result = self.getGuestTestRunResult()
        except HwCertCommandException, e:
            print "Error: could not find guest test results"
            print e
            result = False

        try:
            self.unmountDataImage()
        except:
            print "Warning: could not unmount data image"
            print e

        return result

    def getGuestTestRunResult(self):
        # open the test results, and look for the run
        results = ResultsDocument()
        if not results.load(os.path.join(self.getOutputDirectory(), self.environment.getResultsFileName())):
            print "Error: could not load %s" % os.path.join(self.getOutputDirectory(), self.environment.getResultsFileName())
        print "Loaded %s " % os.path.join(self.getOutputDirectory(), self.environment.getResultsFileName())
        # using recursive version to ignore structural XML change in guest to get all test documents
        tests = results.getTests(names=None, onlyFailures=False, recursive=True)
        if not tests:
            print "Error: no tests in guest results"
            return False
        if self.tagToRun:
            for testDocument in tests:
                print "Checking test %s %s" % (testDocument.getName(), self.getDocDeviceName(testDocument))
                print "  tags: %s" % testDocument.getTags()
            tests = results.getTaggedTests(tests, [self.tagToRun])
            if not tests:
                print "Error: no test tagged " + self.tagToRun + " in guest results"
                return False

        # if the test is run on multiple devices (storage, network), look for one that passed
        for testDocument in tests:
            print "Checking test %s %s" % (testDocument.getName(), self.getDocDeviceName(testDocument))
            # have to check names here to be sure
            if self.testToRun and testDocument.getName() != self.testToRun:
                continue
            for runDocument in testDocument.getRuns():
                print "Run %s %s" % (runDocument.getNumber(), runDocument.getSummary())
                if runDocument.getSummary() == Constants.PASS:
                    if self.enableTargetDevice and self.targetDevice:
                        if self.getDocDeviceName(testDocument) == self.targetDevice:
                            print "Guest test %s on target %s PASSED" % (testDocument.getName(), self.getDocDeviceName(testDocument))
                            return True
                    else: # default target
                        print "Guest test %s PASSED" % testDocument.getName()
                        return True

        # otherwise
        print "Error: " + self.testToRun + " run on guest missing or failed"
        return False

    def getDocDeviceName(self, testDocument):
        """ this is an adapter to read the old pre-rhcert value that was a test attribute  
        it needs to go away when hwcert on the guest is updated to use TestDocument.getTestParameters()"""
        return testDocument.element.getAttribute(Attributes.logical_device)

    def run(self):
        # just pass-through the logical device to the guest call
        if self.getLogicalDeviceName() and self.enableTargetDevice:
            self.targetDevice = self.getLogicalDeviceName()

        # determine whether to use xen or kvm virtualization
        self.configureHypervisor()
        # verify the guest, start if necessary
        if not self.verifyGuest():
            return False

        # submit the tasks to the guest
        if not self.submitTasks():
            return False

        # run the guest, wait for results,
        if not self.runGuest():
            return False

        # then copy the results.xml, and get the guest return value
        if not self.getGuestResults():
            return False

        return True


if __name__ == "__main__":
    test = FvTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
