#!/usr/bin/python
# hal.py - a class for working with HAL.
# Author: Will Woods <wwoods@redhat.com>

# NOTES: 
# This could probably be cleaner but it was really just designed for my own
# convenience. It could also use documentation on the interface for Device, or
# a wrapper class for that.

import dbus, re, string, sys

from rhcert.tags import Constants
from rhcert.command import Command, HwCertCommandException

from device import HalDevice
from deviceDetector import DeviceDetector

class Hal(DeviceDetector):
    "A class for working with HAL."
    def __init__(self):
        DeviceDetector.__init__(self, name=Constants.hal)
        self.debugging = False
        self.enabled = True
        self.__checkServices()
        try:
            self.bus = dbus.SystemBus()
            manager_obj = self.bus.get_object('org.freedesktop.Hal','/org/freedesktop/Hal/Manager')
            self.manager = dbus.Interface(manager_obj,'org.freedesktop.Hal.Manager')
        except Exception, exception:
            if self.debugging:
                if type(exception) is AttributeError:
                    print "Warning: Early DBus API not supported, disabling HAL discovery"
                else:
                    print "Warning: "
                    print exception
            self.enabled = False

    def __checkServices(self):
        for daemon in ["haldaemon", "messagebus"]:
            try:
                Command("service %s status" % daemon).run()
            except HwCertCommandException, e:
                print "Error: HAL daemons \"%s\" is not running - please start it." % daemon
                print e

    def getDevice(self, udi):
        return HalDevice(self.getDeviceProperties(udi))
    def __GetDevice(self,udi):
        "Return a Device object for the given UDI."
        dev_obj=self.bus.get_object('org.freedesktop.Hal',udi)
        return dbus.Interface(dev_obj,'org.freedesktop.Hal.Device')
    def __GetAllDevices(self):
        "Return a list of UDIs for all devices HAL knows about."
        return self.manager.GetAllDevices()
    def getDeviceProperties(self,udi):
        "Return a dict of property info for the device with the given udi."
        return self.__GetDevice(udi).GetAllProperties()
    def getDevices(self):
        if self.enabled:
            "Return a list containing property info dicts for all devices."
            return [HalDevice(self.getDeviceProperties(d)) for d in self.__GetAllDevices()]
        # otherwise
        return list()

    def dumpDevice(self,udi):
        "Return a string containing device properties for the given device"
        dev=self.getDeviceProperties(udi)
        rv="udi = %s\n" % udi
        for (k,v) in [(k,dev[k]) for k in sorted(dev.keys())]:
            if type(v) == type(list()):
                v=[i.encode("utf-8") for i in v]
            rv += "  %s = %s\n" % (k,v)
        return rv
    def dumpall(self):
        "Dump device properties for all devices (like lshal)"
        return "\n".join([self.dumpDevice(udi) for udi in self.__GetAllDevices()])
    def dump(self): return self.dumpall()

    def getUSBBusAndPort(self, device):
        try:
            if not device.getProperty('info.bus') in ('usb',  'usb_device'):
                return None
            # use linux.sysfs_path_device as the id
            deviceID = device.getProperty('linux.sysfs_path')
        except KeyError:
            return None

        # check n-n.n syntax
        pattern = re.compile("\/\d+-\d+\.d+")
        match = pattern.search(deviceID)
        if not match:
            pattern = re.compile("\/\d+-\d+")
            match = pattern.search(deviceID)
        if match: 
            # print "match = %s" % match.group()
            dashPosition = string.find(match.group(), "-")
            busNumber = match.group()[1:dashPosition]
            portNumber = match.group()[dashPosition+1:]
            # print "%s at bus %s port %s" % (deviceID, busNumber, portNumber)
            return (busNumber, portNumber, deviceID)
        return (None, None, deviceID)

    def getPCIBusAndID(self, device):
        """rely on HAL to check PCI info"""
        try:
            # check that this is a pci device (info.bus on RHEL5, info.subsystem on RHEL6)
            if device.getProperty('info.bus') != 'pci' and device.getProperty('info.subsystem') != 'pci':
                return None
        except KeyError:
            return None
        deviceID = None
        try:
            # use linux.sysfs_path_device as the id
            deviceID  = device.getProperty('linux.sysfs_path_device')
        except KeyError:
            pass
        if not deviceID:
            try:
                # use linux.sysfs_path as the id (RHEL 6)
                deviceID  = device.getProperty('linux.sysfs_path')
            except KeyError:
                return None
        # check the detailed pcidom:pcibus:pcidev.pcifn
        deviceInfo = deviceID[-12:]
        if deviceInfo:
            domAndBus = deviceInfo[:7]
            devAndFn  = deviceInfo[8:]
            return (domAndBus, devAndFn, deviceID)
        return (None, None, deviceID)

    def getProduct(self, device):
        return device.getProperty('info.product')

def unit_test():
    try:
        h=Hal()
        print h.dumpall()
        return 0
    except Exception, e:
        print str(e)
        return 1

if __name__ == "__main__":
    sys.exit(unit_test())
