# Copyright (c) 2013 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import subprocess
from rhcert.command import Command, HwCertCommandException

class IWConfig:
    """ parser for the output of the iwconfig command """

    def __init__(self):
        self.interfaces = list()
        self.types = dict() # type[interface]
        self.parameters = dict() # parameter[interface][name]
        self.output = None
        self.__parse()

    def __parse(self, searchInterface=None):
        iwconfig = "iwconfig"
        if searchInterface:
            iwconfig += " %" % searchInterface
        pipe = subprocess.Popen(iwconfig, shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,)
        (output, errors) = pipe.communicate()
        if output:
            self.output = output.splitlines()
            interface = None
            for line in self.output:
                # first line: no wireless extensions | (<interface> <wlan> <parameter list>)
                # following lines: <parameter list>
                # record separator line is blank
                if len(line) == 0:
                    interface = None
                elif not interface:
                    if "no wireless extensions" in line:
                        continue
                    # otherwise
                    interface = line.split("  ")[0]
                    if searchInterface and interface != searchInterface:
                        continue
                    self.interfaces.append(interface)
                    # there's four spaces between interface name and type
                    self.types[interface] = line.split("  ")[2]
                    self.__addParameters(interface, "  ".join(line.split("  ")[3:]))
                else:
                    self.__addParameters(interface, line)


    def __addParameters(self, interface, parameterString):
        # double spaces seem to separate name/value pairs
        nameValues = parameterString.split("  ")
        for nameValue in nameValues:
            # separator is : or =
            pair = nameValue.split(":")
            if "=" in nameValue:
                pair = nameValue.split('=')
            if len(pair) == 2:
                if not interface in self.parameters:
                    self.parameters[interface] = dict()
                self.parameters[interface][pair[0].strip()] = pair[1].strip()

    def getPhyList(self):
        """Get physical device name,
        using rfkill and returning the first device with
        'Wireless LAN' on the same line.
        Give a warning if none is found.
        """
        rfkill_command = Command("rfkill list all")
        phy = ""
        try:
            return rfkill_command.getStringList(regex="^\d: (?P<wifi>.*):.*Wireless LAN", regexGroup="wifi")
        except HwCertCommandException, e:
            print "Warning: no wifi device found"
            return []

    def getType(self):
        """Get wifi interface (ac, n, g, or b),
        using the iw <physical device> info command
        and returning b if none of the criteria are met for the others.
        Return None if there is no physical device.
        """
        phy_list = self.getPhyList()
        for phy in phy_list:
            iw_command = Command("iw %s info" % phy)
            try:
                all_the_info = iw_command.getStringList(regex="^.*")
            except HwCertCommandException, e:
                continue
            vht_capabilities = [x for x in all_the_info if 'VHT Capabilities' in x]
            short_GI80 = [x for x in all_the_info if 'short GI (80 MHz)' in x]
            ht20_ht40 = [x for x in all_the_info if 'HT20/HT40' in x]
            fifty_four_Mbps = [x for x in all_the_info if '54.0 Mbps' in x]

            if vht_capabilities and short_GI80:
                return "ac"
            elif ht20_ht40 and not vht_capabilities:
                return "n"
            elif fifty_four_Mbps and not ht20_ht40 and not vht_capabilities:
                return "g"
            else:
                return "b"
        return None

    def getInterfaces(self): return self.interfaces

    def getParameter(self, interface, parameter):
        if interface in self.parameters and parameter in self.parameters[interface]:
            return self.parameters[interface][parameter]
        # otherwise
        return None

    def dump(self):
        for interface in self.interfaces:
            print "%s %s" % (interface, self.getType())
            for (name, value) in self.parameters[interface].items():
                print "    %s: %s" % (name, value)
