# Copyright (c) 2006-14 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys, time

from rhcert.tags import Constants, DeviceClass, TestTag

from hwcert.hwcert.hardwareTest import HardwareTest
from hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector

class BatteryTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "battery")
        self.acAdapterDevice = None
        self.deviceClass = DeviceClass.battery
        self.unitFactor = 1000 # parse from mu-Wh to mWh
        self.units = "mWh"

    def tags(self):
        return [TestTag.interactive, TestTag.certification, TestTag.portable]

    def setACAdapterDevice(self, device):
        self.acAdapterDevice = device
    def getACAdapterDevice(self):
        return self.acAdapterDevice

    def plan(self, planner):
        tests = list()
        acAdapterDevice = None
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            if "power_supply" in device.getProperty("SUBSYSTEM"):
                if device.getProperty("POWER_SUPPLY_TYPE") == "Battery":
                    test = self.makeCopy()
                    test.setDevice(device)
                    tests.append(test)
                if device.getProperty("POWER_SUPPLY_TYPE") == "Mains":
                    acAdapterDevice = device
        if acAdapterDevice:
            for test in tests:
                test.setACAdapterDevice(acAdapterDevice)
        return tests

    def getDevices(self):
        """ get the battery and ac adapter devices """
        devices = UdevDeviceDetector().getDevices()
        for device in devices:
            if "power_supply" in device.getProperty("SUBSYSTEM"):
                if device.getProperty("POWER_SUPPLY_TYPE") == "Mains":
                    self.setACAdapterDevice(device)
                if device.getProperty("POWER_SUPPLY_TYPE") == "Battery":
                    self.setDevice(device)

        if not self.getDevice():
            print "Error: battery is not present"
        if not self.getDevice() or not self.getACAdapterDevice():
            print "Error: could not find udev device for battery and AC Adapter"
            return False

        return True

    def getStatus(self):
        levelFull = ''
        level = ''
        """ get battery and ac adapter status """
        if self.getDevices():
            self.acAdapterPresent = bool(int(self.getACAdapterDevice().getProperty("POWER_SUPPLY_ONLINE")))
            # fedora uses "ENERGY" instead of "CHARGE"
            levelFull = self.device.getProperty("POWER_SUPPLY_CHARGE_FULL")
            if not levelFull:
                levelFull = self.device.getProperty("POWER_SUPPLY_ENERGY_FULL")
            if levelFull:
                self.levelFull = int(levelFull) / self.unitFactor
            level = self.device.getProperty("POWER_SUPPLY_CHARGE_NOW")
            if not level:
                level = self.device.getProperty("POWER_SUPPLY_ENERGY_NOW")
            self.level = int(level) / self.unitFactor
            self.chargingStatus = self.device.getProperty("POWER_SUPPLY_STATUS").lower()
            return True

        #otherwise
        return False

    def printStatus(self):
        state = "is" if self.acAdapterPresent else "is not"
        print "-------------------------------"
        print "AC Adapter %s connected" % state
        print "Battery:"
        levelPercentage = 100 * self.level / self.levelFull
        print "    charged to %s%% - %s %s" % ( levelPercentage, self.level, self.units)
        print "    current charging status is %s" % self.chargingStatus
        print "-------------------------------"

    def checkStatus(self):
        success = True
        delta = 10     # in units of charge level
        waitTime = 10  # seconds to wait each try
        tryLimit = 10  # number of times to look for level change

        self.getStatus()

        if self.chargingStatus == "discharging":
            if self.acAdapterPresent:
                print "Error: battery is discharging while AC adapter is connected"
                success = False

        if self.chargingStatus == "charging":
            if not self.acAdapterPresent:
                print "Error: battery is charging while AC adapter is disconnected"
                success = False

        # wait if status is unknown:
        tries = 0
        while self.chargingStatus != "charging" and self.chargingStatus != "discharging":
            if tries > tryLimit:
                print "Error: battery charging status is %s after retry limit." % self.chargingStatus
                return False
            # otherwise
            print "battery charging status is %s, waiting..." % self.chargingStatus
            tries = tries + 1
            sys.stdout.flush()
            time.sleep(waitTime)
            self.getStatus()

        # check for changes in battery level.
        tries = 0
        level = self.level
        while True:
            self.getStatus()
            if ((self.chargingStatus == "charging" and self.level > level + delta)
                    or (self.chargingStatus == "discharging" and self.level < level - delta)):
                print "verified battery is %s" % self.chargingStatus
                break
            if tries > tryLimit:
                print "Error: could not verify battery %s" % self.chargingStatus
                success = False
                break
            tries = tries + 1
            print "waiting to verify battery %s more than %s %s" % (self.chargingStatus, delta, self.units)
            sys.stdout.flush()
            time.sleep(waitTime)
        else:
            print "Error: could not verify battery charging status, current status is %s" % self.chargingStatus
            success = False

        return success


    def run(self):
        if self.getMode() == Constants.auto:
            print "Error: battery test can not be run in auto mode, as it requires the tester to disconnect power"
            return False

        if not self.getStatus():
            return False
        self.printStatus()

        action = None
        testedConnected = False
        testedDisconnected = False
        while not testedConnected or not testedDisconnected:
            # ask user to add/remove AC power
            action = "disconnect" if self.acAdapterPresent else "connect"
            if not self.promptConfirm("Please %s AC Power - continue? " % action):
                return False

            if not self.getStatus():
                return False
            if action == "connect":
                if self.acAdapterPresent:
                    testedConnected = True
                else:
                    print "AC Power is not connected!"
                    continue
            elif action == "disconnect":
                if self.acAdapterPresent:
                    print "AC Power is not disconnected!"
                    continue
                else:
                    testedDisconnected = True

            if not self.checkStatus():
                return False

            if not self.getStatus():
                return False
            self.printStatus()

        return True


if __name__ == "__main__":
    test = BatteryTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
