#!/usr/bin/python
# Copyright (c) 2008 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>

import os, string, sys, re, time, decimal, filecmp

from aperfmperf import EffectiveFrequency
from syscpufreq import SysCPUFreq
from decimal import Decimal

from rhcert.tags import Constants, DeviceClass, TestTag, TestResult
from rhcert.command import Command, HwCertCommandException

from hwcert.hwcert.hardwareTest import HardwareTest
from hwcert.hwcert.procDeviceDetector import ProcDeviceDetector

"""
    Tolerance Calculation, workload time checks:
    
    p = predicted speedup = (max freq / min freq)
    s = measured speedup = (workload time, min freq / workload time, max freq)
    t = tolerance

    * CERTPX-238 minimum ignored by setting lower tolerance to 5Ghz
     when 'hwp' cpu flag is present
    
    p - (p - 1.0)*t < s < p + (p - 1.0)
    
    example: p = 2.0, s = 1.9, t = 0.5
    2.0 - (2.0 - 1.0)*0.5 < 1.9 < 2.0 + (2.0-1.0)*.5
    evaluates to:
    1.5 < 1.9 < 2.5  = True (PASS)
"""


class CPUScalingTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "cpuscaling")
        self.speedUpTolerance = 50.0 # percent

        self.lowerTolerance = 50 # MHz
        self.higherTolerance = 100 # MHz
        self.minimumCheckLowerTolerance = 200
        self.minimumCheckHigherTolerance = 200

        self.retryLimit = 5
        self.retryTolerance = 5.0 # percent
        self.sysCPUDirectory = "/sys/devices/system/cpu"
        self.idaFlag = "ida"
        self.hwpFlag = "hwp"
        self.aperfmperfFlag = "aperfmperf"
        self.idaSpeedupFactor = 8.0 # percent
        self.selfTestData = [10.0, 10.0, 10.0, 10.0] # min, max, ondemand, performance
        self.selfTestStep = 0
        self.deviceClass = DeviceClass.processor
        self.cpuNumbers = None
        self.cpuToPackage = None
        self.packageToCpus = None
        self.workloads = ["minimum", "maximum", "ondemand", "performance"]
        self.workloadNames = {"minimum": "User Space (min)", "maximum": "User Space (max)", "ondemand": "On Demand", "performance": "Performance"}
        self.workloadTime = dict() # workload run time list indexed by [workload][cpu]
        self.workloadFreq = dict() # measured cpu freq indexed by [workload][cpu]
        self.workloadEffectiveFreq = dict() # measured effective cpu freq indexed by [workload][cpu]
        self.workloadErrors = dict() # error log indexed by [workload]
        self.currentWorkload  = "minimum"
        self.model = None
        self.maximumModelFrequency = None
        self.sysCPUFreq = SysCPUFreq()
        self.cpuNumbers = None

    def dumpTables(self):
        print "\nDumping Data Tables:"
        for workload in self.workloads:
            print "workload: " + workload
            print "-----------------------------------------------------------------------------------"
            if workload in self.workloadTime:
                for (cpu, times) in self.workloadTime[workload].iteritems():
                    sys.stdout.write("cpu%s: " % cpu)
                    for time in times:
                        sys.stdout.write("  %.2fs" % time)
                    print ""
            if workload in self.workloadFreq:
                for (cpu, freqs) in self.workloadFreq[workload].iteritems():
                    sys.stdout.write("cpu%s: " % cpu)
                    for freq in freqs:
                        sys.stdout.write("  %4u MHz" % freq)
                    print ""
            if workload in self.workloadEffectiveFreq:
                for (cpu, freqs) in self.workloadEffectiveFreq[workload].iteritems():
                    sys.stdout.write("cpu%s: " % cpu)
                    for freq in freqs:
                        sys.stdout.write("  %4u MHz" % freq)
                    print ""

    def setWorkloadTime(self, workload, cpu, time):
        if not workload in self.workloadTime:
            self.workloadTime[workload] = dict()
        if not cpu in self.workloadTime[workload]:
            self.workloadTime[workload][cpu] = list()
        self.workloadTime[workload][cpu].append(time)

    def setWorkloadFreq(self, workload, cpu, freq):
        if not workload in self.workloadFreq:
            self.workloadFreq[workload] = dict()
        if not cpu in self.workloadFreq[workload]:
            self.workloadFreq[workload][cpu] = list()
        mhz = int(freq)/1000
        self.workloadFreq[workload][cpu].append(mhz)

    def setWorkloadEffectiveFreq(self, workload, cpu, freq):
        if not workload in self.workloadEffectiveFreq:
            self.workloadEffectiveFreq[workload] = dict()
        if not cpu in self.workloadEffectiveFreq[workload]:
            self.workloadEffectiveFreq[workload][cpu] = list()
        mhz = int(freq)/1000
        self.workloadEffectiveFreq[workload][cpu].append(mhz)

    def logWorkloadError(self, errorMessage):
        print errorMessage
        if not self.currentWorkload in self.workloadErrors:
            self.workloadErrors[self.currentWorkload] = list()
        self.workloadErrors[self.currentWorkload].append(errorMessage)

    def getWorkloadTime(self, workload, cpu):
        try:
            return self.workloadTime[workload][cpu][-1] # get the last time
        except KeyError:
            return None

    def getWorkloadFreq(self, table, workload, cpu):
        try:
            return table[workload][cpu][-1] # get the last freq
        except KeyError:
            return None

    def getWorkloadTimes(self, workload, cpu):
        try:
            return self.workloadTime[workload][cpu] # get the whole list of times
        except KeyError:
            return None
    def getWorkloadTimesLength(self, workload):
        """ return the max number of entries of any cpu """
        max = 0
        for cpu in self.workloadTime[workload]:
            if len(self.workloadTime[workload][cpu]) > max:
                max = len(self.workloadTime[workload][cpu])
        return max

    def speedup(self, cpu, workload="maximum"):
        try:
            return self.getWorkloadTime("minimum", cpu)/self.getWorkloadTime(workload, cpu)
        except Exception, e:
            print "Error: could not determine speedup for workload %s" % workload
            print e
            return None

    def checkFrequency(self, expected, actual, lowerTolerance=None, higherTolerance=None):
        """Check that expected-lowerTolerance <= actual <= expected+higherTolerance

        >>> c_test = CPUScalingTest()
        >>> c_test.lowerTolerance = 50
        >>> c_test.higherTolerance = 100
        >>> c_test.checkFrequency(100*1000, 12*1000)
        False
        >>> c_test.checkFrequency(100*1000, 50*1000)
        True
        >>> c_test.checkFrequency(100*1000, 201*1000)
        False
        >>> c_test.checkFrequency(100*1000, 200*1000)
        True
        """

        if not lowerTolerance:
            lowerTolerance = self.lowerTolerance

        if not higherTolerance:
            higherTolerance = self.higherTolerance

        # expected and actual int, in KHz
        lowest_tolerated = int(expected) - lowerTolerance*1000
        highest_tolerated = int(expected) + higherTolerance*1000
        return lowest_tolerated <= int(actual) and highest_tolerated >= int(actual)

    def checkFrequencyGreaterThan(self, freq1, freq2):
        # freq1 and freq2 is strings, in KHz
        # check if freq1+tolerance >= freq2
        return int(freq1) + self.higherTolerance*1000 > int(freq2)

    def checkSpeedup(self, speedup, warn=False, maximumLimit=True):
        if not self.predictedSpeedup:
            self.logWorkloadError("Error: could not predict speedup, unable to evaluate measured speedup")
            return False

        message = None
        tag = "Error:"
        if warn:
            tag = "Warning:"

        if speedup < self.minimumRequiredSpeedup():
            message = "%s measured speedup %.2f less than the minimum speedup of %.2f. " % (tag, speedup, self.minimumRequiredSpeedup())
        if speedup > self.maximumRequiredSpeedup():
            # if no maximumLimit, downgrade to warn
            if not maximumLimit:
                warn = True
                tag = "Warning:"
            message = "%s measured speedup %.2f greater than the maximum speedup of %.2f" % (tag, speedup, self.maximumRequiredSpeedup())

        if message:
            if warn:
                print message
            else:
                self.logWorkloadError(message)
                return False

        return True

    def minimumRequiredSpeedup(self):
        return self.predictedSpeedup - ((self.predictedSpeedup - 1.00)*self.speedUpTolerance/100.0)

    def maximumRequiredSpeedup(self):
        return self.predictedSpeedup + ((self.predictedSpeedup - 1.00)*self.speedUpTolerance/100.0)

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        # check the /sys file system to see if scaling looks possible
        tests = list()
        #for x86 architectures, check if C-state is supported if not check if P-state is supported
        if self.getCPUarch().find("x86") == 0:
            if self.sysCPUFreq.isCStateSupported() or self.sysCPUFreq.isFrequencyScalingSupported():
                test = self.makeCopy()
                tests.append(test)
        elif self.sysCPUFreq.isFrequencyScalingSupported():
            test = self.makeCopy()
            tests.append(test)
        return tests

    def getRequiredRPMs(self):
        return ["cpupowerutils"]

    def verify(self):
        # disable for now
        return True

    def logDevices(self, planner):
        devices = list()
        proc = ProcDeviceDetector()
        devices.extend(proc.getDevices(ProcDeviceDetector.procCPUInfo))
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def getCPUarch(self):
        self.cpuArch = None
        try:
            self.cpuArch = Command("uname -m").getString()
        except HwCertCommandException, exception:
            print "Warning: could not find the CPU architecture"
            print exception
        return self.cpuArch

    def getCPUFreqDirectories(self):
        if not os.path.exists(self.sysCPUDirectory):
            print "Error: no file %s" % self.sysCPUDirectory
            return None
        # look for cpu subdirectories
        pattern = re.compile("cpu(?P<cpuNumber>[0-9]+)")
        self.cpufreqDirectories = None
        for subdirectory in os.listdir(self.sysCPUDirectory):
            match = pattern.search(subdirectory)
            if match and match.group("cpuNumber"):
                cpuNumber = match.group("cpuNumber")
                if not self.cpuNumbers:
                    self.cpuNumbers = list()
                self.cpuNumbers.append(cpuNumber)
                cpufreqDirectory = "%s/%s/cpufreq" % (self.sysCPUDirectory, subdirectory)
                if not os.path.exists(cpufreqDirectory):
                    continue
                # otherwise
                if not self.cpufreqDirectories:
                    self.cpufreqDirectories = dict()
                self.cpufreqDirectories[cpuNumber] = cpufreqDirectory
        return self.cpufreqDirectories

    def checkParameters(self, file):
        current = None
        for cpufreqDirectory in self.cpufreqDirectories.values():
            parameters = self.getParameters(cpufreqDirectory, file)
            if not parameters:
                print "Error: could not determine cpu parameters from %s/%s" % (cpufreqDirectory, file)
                return None
            # otherwise
            if not current:
                current = parameters
            elif not current == parameters:
                return None

        return current

    def getParameters(self, cpufreqDirectory, file):
        try:
            file = open("%s/%s" % (cpufreqDirectory, file))
            while 1:
                line = file.readline()
                if not line:
                    break
                if len(line.strip()) > 0:
                    return line.strip().split()
        except Exception, e:
            if self.getDebug() != Constants.off:
                print e
        return None

    def setPackageParameter(self, switch, setFile, readFile, value, verify=True):
        # try the command for all cpus in the package
        result = True
        for cpu in self.packageToCpus[self.currentPackage]:
            if not self.setParameter(cpu, switch, setFile, readFile, value, verify):
                result = False
        return result

    def setParameter(self, cpu, switch, setFile, readFile, value, verify=True):
        result = True
        try:
            command = Command("cpupower -c %s frequency-set -%s %s" % (cpu, switch, value))
            command.echo()
        except HwCertCommandException, exception:
            print "Note: command failed: %s" % exception
            command.printErrors()
            result = False

        if not result:
            print "Trying alternate method."
            try:
                command = Command("echo \"%s\" > %s/%s" % (value, self.cpufreqDirectories[cpu],  setFile))
                command.echo()
            except HwCertCommandException, exception:
                print "Error: command failed:"
                print exception
                return False

        # verify it has changed
        if verify:
            parameterFile = open("%s/%s" % (self.cpufreqDirectories[cpu], readFile))
            line = parameterFile.readline()
            if not line or line.strip() != value:
                self.logWorkloadError("Error: could not verify that %s/%s was set to %s" % (self.cpufreqDirectories[cpu], readFile, value))
                if line:
                    print "Actual Value: %s" % line
                else:
                    print "parameter file was empty"
                return False

        return True

    def setFrequency(self, frequency):
        if self.setPackageParameter("f", "scaling_setspeed", "scaling_cur_freq", frequency, verify=False):
            return self.getFrequency()

        # otherwise, set command has failed
        return None

    def getFrequency(self, cpu=None):
        if not cpu:
            print "\nChecking current cpu frequency:"
            cpuInfoCurrentFreq = self.getPackageParameter("cpuinfo_cur_freq", logging=True)
            scalingCurrentFreq = self.getPackageParameter("scaling_cur_freq", logging=True)
            # if cpuInfoCurrentFreq != scalingCurrentFreq:
            #    print "Warning: scaling_cur_freq is %s while cpuinfo_cur_freq is %s" % ( scalingCurrentFreq, cpuInfoCurrentFreq)
            if cpuInfoCurrentFreq:
                print "Using cpuinfo_cur_freq"
                return cpuInfoCurrentFreq

            self.logWorkloadError("Error: could not determine current package frequency")
            return None

        # otherwise - single cpu
        print "\nChecking current frequency for cpu%s" % cpu
        cpuInfoCurrentFreq = self.getParameter(cpu, "cpuinfo_cur_freq", logging=True)
        scalingCurrentFreq = self.getParameter(cpu, "scaling_cur_freq", logging=True)
        if cpuInfoCurrentFreq:
            print "Using cpuinfo_cur_freq"
            return cpuInfoCurrentFreq

        self.logWorkloadError("Error: could not determine current frequency")
        return None


    def setGovernor(self, governor):
        return self.setPackageParameter("g", "scaling_governor", "scaling_governor", governor)

    def getPackageParameter(self, parameter, logging=False):
        packageParameterValue = None
        for cpu in self.packageToCpus[self.currentPackage]:
            value = self.getParameter(cpu, parameter)
            if logging:
                print "cpu%s %s = %s" % (cpu, parameter, value)
            if value and not packageParameterValue:
                packageParameterValue = value
            elif value != packageParameterValue:
                print "Warning: cpu%s in package %s has the value %s which differs from other cpus in the package" % (cpu, self.currentPackage, value)

        return packageParameterValue

    def getParameter(self, cpu, parameter, logging=False):
        value = None
        parameterFilePath = "%s/%s" % (self.cpufreqDirectories[cpu], parameter)
        try:
            parameterFile = open(parameterFilePath)
            line = parameterFile.readline()
            if not line:
                self.logWorkloadError("Error: failed to get %s for %s" % (parameter, self.cpufreqDirectories[cpu]))
                return None
            value = line.strip()
            if logging:
                print "cpu%s %s = %s" % (cpu, parameter, value)
            return value
        except IOError, exception:
            self.logWorkloadError("Error: could not open %s" % parameterFilePath)
            print exception

        return None

    def getParameterList(self, parameter):
        values = list()
        for cpufreqDirectory in self.cpufreqDirectories.values():
            parameterFile = open("%s/%s" % (cpufreqDirectory, parameter))
            line = parameterFile.readline()
            if not line:
                self.logWorkloadError("Error: failed to get %s for %s" % (parameter, cpufreqDirectory))
                return None
            values.append(line.strip())

        return values

    def runLoadTest(self, cpu=None):
        """ run package load test, cpu default None means for all cpus """
        print "Running CPU load test..."

        runTime = None
        tries = 0
        while tries < self.retryLimit:
            sys.stdout.flush()
            averageProcessTime = self.__runSubprocessLoadTest(cpu)
            if averageProcessTime:
                print "average worker process time: %.2f seconds" % averageProcessTime
                if not runTime:
                    runTime = averageProcessTime
                else:
                    thisTime = averageProcessTime
                    # print "comparing %.2f" % (abs(thisTime-runTime)/runTime)
                    if (abs(thisTime-runTime)/runTime)*100 < self.retryTolerance:
                        return thisTime
                    else:
                        runTime = thisTime
            tries += 1
            # print "try %u - %.2f sec" % (tries, runTime)

        print "Warning: could not repeat load test times within %.1f%%" % self.retryTolerance
        return runTime

    def __runSubprocessLoadTest(self, singleCpu=None):
        print "Running load test for package " + self.currentPackage
        self.loadProcesses = dict()

        # if cpu is set, just load that one
        if singleCpu != None:
            loadProcess = LoadProcess(self, singleCpu, aperfmperfSupport=self.effectiveFrequencySupport)
            loadProcess.start()
            self.loadProcesses[singleCpu] = loadProcess
        else: # load all cpus in the package
            for cpu in self.packageToCpus[self.currentPackage]:
                loadProcess = LoadProcess(self, cpu, aperfmperfSupport=self.effectiveFrequencySupport)
                loadProcess.start()
                self.loadProcesses[cpu] = loadProcess

        print "waiting for load processes..."
        totalProcessTime = 0
        numberOfProcesses = len(self.loadProcesses)
        while len(self.loadProcesses) > 0 :
            for (cpu, process) in self.loadProcesses.items():
                if process.isDone():
                    if process.returnCode != 0:
                        self.logWorkloadError("Error: load process terminated with return code %s" % process.returnCode)
                        return None
                    runTime = process.getRunTime()
                    frequency = self.getFrequency(cpu)
                    effectiveFrequency = process.getEffectiveFrequency()
                    if not runTime:
                        self.logWorkloadError("Error: could not find process run time")
                        return None
                    if not frequency:
                        self.logWorkloadError("Error: could not determine cpu frequency")
                        return None
                    # otherwise - collect info
                    print "process for cpu %s is done in %.2f seconds, at %u MHz" % (cpu, runTime, int(int(frequency)/1000))
                    self.setWorkloadTime(self.currentWorkload, cpu, runTime)
                    self.setWorkloadFreq(self.currentWorkload, cpu, frequency)
                    if effectiveFrequency:
                        print "process effective frequency: %u MHz" % (effectiveFrequency/1000)
                        self.setWorkloadEffectiveFreq(self.currentWorkload, cpu, effectiveFrequency)
                    else:
                        print "Warning: could not determine effective frequency"
                    totalProcessTime += runTime
                    del self.loadProcesses[cpu]

            time.sleep(1) # poll once per second
            # if doing a single cpu, log its current freq while waiting
            if singleCpu != None:
                runningFreq = self.getFrequency(cpu)

        print "processes complete"
        return totalProcessTime/numberOfProcesses

    def done(self, cpu):
        print "process for cpu %s is done. (caught SIGCHLD)" % cpu
        try:
            del self.loadProcesses[cpu]
        except KeyError:
            self.logWorkloadError("Error: no cpu %s in load processes" % cpu)

    def verifyMinimumFrequency(self):
        waitTime = 5
        sys.stdout.write("Waiting %d seconds..." % waitTime)
        sys.stdout.flush()
        time.sleep(waitTime)
        sys.stdout.write(" done.\n")
        minimumFrequency = self.getPackageParameter("scaling_min_freq")
        currentFrequency = self.getFrequency()
        if not minimumFrequency or not currentFrequency or not self.checkFrequency(minimumFrequency, currentFrequency, self.minimumCheckLowerTolerance, self.minimumCheckHigherTolerance):
            # CERTPX-238
            if self.hwpFlag not in self.cpuFlags:
                return False

        # otherwise
        return True

    def getSystemCapabilities(self):
        print ""
        print "System Capabilites:"
        print "-------------------------------------------------"

        self.logCPUInfo()

        # 1.  Determine if the cpu's support scaling using presense or lack there of ~/cpufreq in sysfs

        if not self.getCPUFreqDirectories():
            print "Error: Could not find frequency directories."
            return False
        if len (self.cpufreqDirectories) > 1:
            print "System has %u cpus"% len(self.cpufreqDirectories)

        # 1.5 get core/package topology
        if self.cpuNumbers:
            for cpuNumber in self.cpuNumbers:
                sys.stdout.flush()
                parameters = self.getParameters(os.path.join(self.sysCPUDirectory, ("cpu" + cpuNumber), "topology"), "physical_package_id")
                if parameters and len(parameters) == 1:
                    packageNumber = parameters[0]
                    if not self.cpuToPackage:
                        self.cpuToPackage = dict()
                    self.cpuToPackage[cpuNumber] = packageNumber
                    if not self.packageToCpus:
                        self.packageToCpus = dict()
                    if not packageNumber in self.packageToCpus:
                        self.packageToCpus[packageNumber] = list()
                    if cpuNumber not in self.packageToCpus[packageNumber]:
                        self.packageToCpus[packageNumber].append(cpuNumber)
                else:
                    print "Warning: could not determine cpu topology: no package ID for cpu " + cpuNumber
            # print them out
            if not self.packageToCpus:
                return False
            for package in self.packageToCpus.keys():
                sys.stdout.write("package %s has cpus: " % package)
                sys.stdout.write(", ".join(self.packageToCpus[package]))
                print ""
            print ""

        # 2. Check ~/cpu[*]/scaling_available_freq files to ensure that all cpu's scale the same amount; if not bail
        print "Checking freguencies"
        sys.stdout.flush()
        freqFileName = "scaling_available_frequencies"
        frequencies = self.checkParameters(freqFileName)
        if not frequencies:
            return False
        # sort them as integers
        self.frequencies = map(string.atoi, frequencies)
        self.frequencies.sort()

        print ""
        print "Supported CPU Frequencies: "
        for freq in self.frequencies:
            print "    %u MHz" % (freq/1000)
        print ""

        # 3. Check ~/cpu[*]/scaling_available_governors files to verify all cpu's support the same control methods
        governorFileName = "scaling_available_governors"
        self.governors = self.checkParameters(governorFileName)
        if not self.governors:
            return False

        # 4. Log the testable govs. which I suggest to be "ondemand", "userspace", and "performance"; each test below is as applicable
        print ""
        print "Supported Governors: "
        comma = False
        for governor in self.governors:
            print "    %s" % governor

        self.originalGovernors = self.getParameterList("scaling_governor")
        if self.originalGovernors:
            print ""
            print "Current governors:"
            i = 0
            for g in self.originalGovernors:
                sys.stdout.write("    cpu%u: %s\n" % (i, g))
                i += 1
        else:
            self.logWorkloadError("Error: could not determine current governor settings")
            return False

        print "\nCPU Flags:"
        self.getCPUFlags()
        self.turboBoostSupport = False
        support = " NOT"
        if self.idaFlag in self.cpuFlags:
            support = ""
            self.turboBoostSupport = True
        print "    %s: Turbo Boost is%s supported" % (self.idaFlag, support)
        support = " NOT"
        if self.aperfmperfFlag in self.cpuFlags:
            support = ""
        print "    %s: aperf/mperf is%s supported"% (self.aperfmperfFlag, support)
        sys.stdout.flush()

        print "\n Measurement Method:"
        # use ./aperfmperf.py to assess effective frequency support, as cpuflags are not reliable.
        efi = EffectiveFrequency(0)
        if efi.has_support():
            self.effectiveFrequencySupport = True
            print "   /dev/cpu/0/cpuid indicates aperf/mperf is supported"
            print "   Using C sqrt load processes and aperf/mperf effective freq. measurement (aperf.c)"
        else:
            self.effectiveFrequencySupport = False
            print "   /dev/cpu/0/cpuid indicates aperf/mperf is NOT supported"
            print "   Using Python pi load processes with /sys freq. values, (no aperf/mperf support)."
        print ""
        sys.stdout.flush()

        return True

    def getCPUFlags(self):
        # for now, assume all cpus have the same flags.
        self.cpuFlags = None
        try:
            flagsCommand = Command("cat /proc/cpuinfo")
            self.cpuFlags = flagsCommand.getString(regex="^flags[^:]+:(?P<flags>.*)", regexGroup="flags", singleLine=False)
        except HwCertCommandException, exception:
            print "Warning: could not read CPU flags"
            print exception

    def logCPUInfo(self):
        # determine cpu vendor
        proc = ProcDeviceDetector()
        self.procCPUinfoCPUs = proc.getDevices(ProcDeviceDetector.procCPUInfo)
        if self.procCPUinfoCPUs:
            self.model = self.procCPUinfoCPUs[0].getProperty("model name")

            # check /proc/cpuinfo model string for maxFreq
            maxFreq = None
            pattern = re.compile("(?P<max>[0-9\.]+)(?P<units>(MHz|GHz))")
            match = pattern.search(self.model)
            if match:
                maxFreq = int(Decimal(match.group("max"))*1000000) # to KHz
                if match.group("units") == "MHz":
                    maxFreq = int(maxFreq/1000)
            if maxFreq:
                self.maximumModelFrequency = maxFreq
        # log the cpus
        try:
            cpuinfo = Command("cat /proc/cpuinfo")
            cpuinfo.run()
            model = None
            cpu = None
            for line in cpuinfo.output:
                if line.startswith("processor"):
                    cpu = line.split(":")[1]
                if line.startswith("model name"):
                    thisModel = line.split(":")[1]
                    if model != thisModel:
                        print "CPU %s Model: %s" % (cpu, thisModel)
                        model = thisModel
        except Exception, exception:
            print "Warning: could not read CPU Info"
            print exception

        return True

    def getMaximumFrequency(self):
        if self.maximumModelFrequency:
            return self.maximumModelFrequency
        # otherwise
        return self.getPackageParameter("scaling_max_freq")

    def turboBoostSupported(self):
        return self.turboBoostSupport

    def checkFrequencyResults(self, requiredFrequency, unlimitedMaximum=False, lowerTolerance=None, higherTolerance=None):
        """ requiredFrequency: string, in KHz """

        if not lowerTolerance:
            lowerTolerance = self.lowerTolerance

        if not higherTolerance:
            higherTolerance = self.higherTolerance

        success = True
        if self.effectiveFrequencySupport:
            freqTable = self.workloadEffectiveFreq
        else: # use /sys values
            freqTable = self.workloadFreq

        for cpu in self.packageToCpus[self.currentPackage]:
            frequencyMHz = self.getWorkloadFreq(freqTable, self.currentWorkload, cpu)
            if not frequencyMHz:
                self.logWorkloadError("Error: could not determine the measured frequency for cpu %s under workload %s" % (cpu, self.currentWorkload))
                success = False
                continue
            frequency = frequencyMHz*1000 # KHz
            if unlimitedMaximum:
                if not self.checkFrequencyGreaterThan(frequency, requiredFrequency):
                    self.logWorkloadError("Error: cpu%s in package %s has a measured frequency of %u MHz vs. a requirement of %u MHz (+ %u MHz tolerance)"
                                            % (cpu, self.currentPackage, int(frequency)/1000, int(requiredFrequency)/1000, higherTolerance))
                    success = False
            else:
                if not self.checkFrequency(requiredFrequency, frequency, lowerTolerance, higherTolerance):
                    self.logWorkloadError("Error: cpu%s in package %s has a measured frequency of %u MHz vs. a requirement of %u MHz - %u/+ %u MHz"
                                            % (cpu, self.currentPackage, int(frequency)/1000, int(requiredFrequency)/1000, lowerTolerance, higherTolerance))
                    success = False

        return success

    def runUserSpaceTests(self):
        print ""
        print "On Userspace Governor Test:"
        print "-------------------------------------------------"
        sys.stdout.flush()
        # 5. Set the govr.(*) to "userspace" for each cpu via a command similar to:
        # 6. Verify the govr. setting took by checking ~/scaling_governor
        self.differenceSpeedUp = None

        success = True
        governor = "userspace"
        if governor in self.governors:
            print "Setting governor to %s" % governor
            if not self.setGovernor(governor):
                success = False

            # 7. Set the the cpu speed to it's lowest value
            frequency = self.frequencies[0]
            currentFrequency = self.getFrequency()
            if currentFrequency:
                print "Changing cpu frequency from %u to %u MHz" % (int(currentFrequency)/1000, (frequency/1000))
            else:
                print "Setting cpu frequency to %u MHz" % (frequency/1000)
            if not self.setFrequency("%u" % frequency):
                success = False
            self.currentWorkload = "minimum"
            sys.stdout.flush()

            # 8. Verify the speed is set to the lowest value by comparing ~/scaling_min_freq to ~/scaling_cur_freq
            # if not, it's a warning - the frequency will be checked again after running the load.
            currentFrequency = self.getFrequency()
            minimumFrequency = self.getPackageParameter("scaling_min_freq")
            if not minimumFrequency or not currentFrequency or not self.checkFrequency(minimumFrequency, currentFrequency, self.minimumCheckLowerTolerance, self.minimumCheckHigherTolerance):
                print "Warning: Could not verify that cpu frequency is set to the minimum value of %s" % minimumFrequency
            # calculate predicted speedup for the package
            maximumFrequency = self.getMaximumFrequency()
            try:
                self.predictedSpeedup = float(maximumFrequency)/float(minimumFrequency)
            except Exception, e:
                print "Error: could not determine predicted speedup"
                print e
                self.predictedSpeedup = None
                success = False

            # 9. Run and time non-cacheable work(?) test for each CPU
            # 10. Repeat and time until execution times are within self.retryTolerance of each other
            # run to verify the base line execution time, bail if self.retryTolerance margin cannot be acheived in self.retryLimit runs.
            self.minimumFrequencyTestTime = self.runLoadTest()
            if not self.minimumFrequencyTestTime:
                return False
            print "Minumum frequency average load test time: %.2f" % self.minimumFrequencyTestTime

            # 10.5 check that measured frequency is near minimum
            if not self.checkFrequencyResults(minimumFrequency, self.minimumCheckLowerTolerance, self.minimumCheckHigherTolerance):
                success = False

            # 11. Set the cpu speed to it's highest value as above.
            frequency = self.frequencies[-1]
            currentFrequency = self.getFrequency()
            print "Changing cpu frequency from %u to %u MHz" % (int(currentFrequency)/1000, (frequency/1000))
            if not self.setFrequency("%u" % frequency):
                success = False
            self.currentWorkload = "maximum"

            currentFrequency = self.getFrequency()
            if not maximumFrequency or not currentFrequency or not self.checkFrequency(maximumFrequency, currentFrequency):
                print "Warning: Could not verify that cpu frequency is set to the maximum value of %s" % maximumFrequency

            # 12. Repeat workload test, record timing
            self.maximumFrequencyTestTime = self.runLoadTest()
            if not self.maximumFrequencyTestTime:
                return False
            print "Maximum frequency average load test time: %.2f" % self.maximumFrequencyTestTime

            # 12.5 check the measured max measured freqency is >= stated max
            if not self.checkFrequencyResults(maximumFrequency, unlimitedMaximum=True):
                success = False

            # 13. Verify MHz increase is comparable to time % decrease( eg. slow MHz/fast MHz ~= fast time/slow time;
            # again with self.speedUpTolerance margin)

            # for each cpu in the package
            for cpu in self.packageToCpus[self.currentPackage]:
                measuredSpeedup = self.speedup(cpu)
                print ""
                print "CPU Frequency Speed Up: %.2f" % self.predictedSpeedup
                print "CPU %s Measured Speed Up: %.2f" % (cpu, measuredSpeedup)
                self.checkSpeedup(measuredSpeedup, warn=True)
        else:
            print "Note: %s governor not supported" % governor

        return success

    def runPackageOnDemandTests(self):
        """ NOTE: whole-package ondemand testing is currently unused by the test """
        print ""
        print "On Demand Governor Test:"
        print "-------------------------------------------------"
        self.differenceOnDemandVsMaximum = None
        self.onDemandTestTime = None
        governor = "ondemand"
        success = True
        if governor in self.governors:
            # 14. Set the govr. to "ondemand"
            #15. Verify the govr. is set as above
            print "Setting governor to %s" % governor
            if not self.setGovernor(governor):
                success = False
            self.currentWorkload = "ondemand"

            # 16. Wait a fixed period of time, then verify current speed is the slowest in as before
            if not self.verifyMinimumFrequency():
                self.logWorkloadError("Error: Could not verify that cpu frequency has settled to the minimum value")
                success = False

            # 17. Repeat workload test, record timing; also during testing verify the ~scaling_cur_freq reaches ~scaling_max_freq
            self.onDemandTestTime = self.runLoadTest()
            if not self.onDemandTestTime:
                return False
            print "On Demand load test time: %.2f" % self.onDemandTestTime

            # 18. Compare the timing to the max results from earlier, again time should be within self.speedUpTolerance
            # for each cpu in the package
            for cpu in self.packageToCpus[self.currentPackage]:
                if self.getWorkloadTime("maximum", cpu):
                    onDemandSpeedup = self.speedup(cpu, "ondemand")
                    print "CPU %s Speedup On Demand: %.2f" % (cpu, onDemandSpeedup)
                    if not self.checkSpeedup(onDemandSpeedup, warn=False, maximumLimit=False):
                        success = False
                else:
                    print "No maximum frequency test data to compare"
                    success = False

            # 19. Verify the current speed has returned to the lowest speed as before
            if not self.verifyMinimumFrequency():
                self.logWorkloadError("Error: Could not verify that cpu frequency has settled to the minimum value")
                success = False
        else:
            print "Note: %s governor not supported" % governor

        return success

    def runPerformanceTests(self):

        print ""
        print "Performance Governor Test:"
        print "-------------------------------------------------"
        self.differencePerformanceVsMaximum = None
        # 20. Set and verify the govr. to "performance"
        governor = "performance"
        success = True
        if governor in self.governors:
            print "Setting governor to %s" % governor
            if not self.setGovernor(governor):
                success = False
            self.currentWorkload = "performance"

            # 21. Verify the current speed is the same as scaling_max_freq
            maximumFrequency = self.getMaximumFrequency()
            currentFrequency = self.getFrequency()

            if not maximumFrequency or not currentFrequency or not self.checkFrequency(maximumFrequency, currentFrequency):
                print "Warning: Current cpu frequency of %s is not set to the maximum value of %s" % (currentFrequency, maximumFrequency)

            # 22. Repeat workload test, record timing, also verify frequency does not drop during run
            self.performanceTestTime = self.runLoadTest()
            if not self.performanceTestTime:
                return False
            # otherwise
            print "Performance load test time: %.2f" % self.performanceTestTime

            # 22.5 check EFI freq vs. max
            if not self.checkFrequencyResults(maximumFrequency, unlimitedMaximum=True):
                success = False

            # 23. Compare the timing to the max results for a self.speedUpTolerance delta
            for cpu in self.packageToCpus[self.currentPackage]:
                performanceSpeedup = self.speedup(cpu, "performance")
                print "CPU %s Performance Speedup: %.2f" % (cpu, performanceSpeedup)
                self.checkSpeedup(performanceSpeedup, warn=True)
        else:
            print "Note: %s governor not supported" % governor

        return success

    def runOnDemandTests(self):
        """ test one cpu in the package while others are idle, and verify that its speedup is greater than when all
            cpus in the package are loaded """
        print ""
        print "Single CPU On Demand Test:"
        print "-------------------------------------------------"
        # 30. Set and verify the govr. to "performance"
        governor = "ondemand"
        success = True
        if governor in self.governors:
            print "Setting governor to %s" % governor
            if not self.setGovernor(governor):
                success = False
            self.currentWorkload = "ondemand"

            # 16. Wait a fixed period of time, then verify current speed is the slowest in as before
            if not self.verifyMinimumFrequency():
                print "Warning: Could not verify that cpu frequency has settled to the minimum value"

            # 32 Pick a cpu other than cpu0 to load
            singleCpu = self.getSingleCpu()
            print "Using cpu %s to test Single CPU On Demand." % singleCpu

            # 33. Repeat workload test, loading a single cpu
            self.onDemandTime = self.runLoadTest(cpu=singleCpu)
            if not self.onDemandTime:
                return False
            # otherwise
            print "Single CPU load test time: %.2f" % self.onDemandTime

            # 23. Compare the timing to the max results for a self.speedUpTolerance delta
            ondemandSpeedup = self.speedup(singleCpu, "ondemand")
            if not ondemandSpeedup:
                return False
            print "Single CPU On Demand Speedup: %.1f" % ondemandSpeedup

            if not self.checkSpeedup(ondemandSpeedup, warn=False, maximumLimit=False):
                success = False

        else:
            # some older systems don't support ondemand at all.
            print "Warning: %s governor not supported" % governor
            success = True

        return success

    def getSingleCpu(self):
        # 32 Pick a cpu other than cpu0 to load
        for cpu in self.packageToCpus[self.currentPackage]:
            if cpu != "0":
                break
        return cpu

    def restoreGovernors(self):
        print "Restoring original governor to %s" % (self.originalGovernors[0])
        self.setGovernor(self.originalGovernors[0])

    def runSummary(self):
        for package in self.packageToCpus.keys():
            self.printSummary(package)

        if False: # disable for now:
            for package in self.packageToCpus.keys():
                self.printWorkloadTimeSummary(package)

        if self.workloadErrors:
            print "\nError Summary:"
            print "----------------"
            for workload in self.workloads:
                if workload in self.workloadErrors:
                    for error in self.workloadErrors[workload]:
                        print "%s: %s" % (self.workloadNames[workload], error)
            return False

        return True


    def printSummary(self, package):

        if self.getDebug() != Constants.off:
            self.dumpTables()

        if self.effectiveFrequencySupport:
            freqTable = self.workloadEffectiveFreq
            self.printFreqSummary("CPU Effective Frequency", package, freqTable)
        else:
            freqTable = self.workloadFreq
            self.printFreqSummary("CPU Frequency Test (No Effective Frequency Support)", package, freqTable)


        if "ondemand" in self.workloadTime:
            print "\nCPU Workload Test:\n"
            print "Expected Speedup: %.2f" % self.predictedSpeedup
            print "Allowable Speedup: %.2f to %.2f" % (self.minimumRequiredSpeedup(), self.maximumRequiredSpeedup())
            print ""
            print "           On Demand "
            print "-------- -------------- "

            #actual values, one row per cpu
            workload = "ondemand"
            for cpu in self.packageToCpus[package]:
                workloadTime = self.getWorkloadTime(workload, cpu)
                if workloadTime:
                    sys.stdout.write("cpu %-3s  " % cpu)
                    sys.stdout.write(" %.2f" % self.speedup(cpu, workload))
                    sys.stdout.write((" (%4uMHz, %.2fs)      " % (self.getWorkloadFreq(freqTable, workload, cpu) ,workloadTime)[:10]))
                    print ""

            print ""

    def printFreqSummary(self, title, package, freqTable):
        print ""
        print "Summary for Package %s:" % package
        print "\n%s:\n" % title
        print "           User Min       User Max      Performance"
        print "-------- -------------- -------------- --------------"

        # expected values
        minimumFrequency = int(self.getPackageParameter("scaling_min_freq"))/1000
        maximumFrequency = int(self.getMaximumFrequency())/1000
        print "expected %5u MHz      %5u MHz      %5u MHz" % (minimumFrequency, maximumFrequency, maximumFrequency)

        #actual values, one row per cpu
        for cpu in self.packageToCpus[package]:
            sys.stdout.write("cpu %-3s  " % cpu)
            for workload in self.workloads:
                if workload in self.workloadTime and workload != "ondemand" and workload != "turbo":
                    freq = self.getWorkloadFreq(freqTable, workload, cpu)
                    if freq:
                        sys.stdout.write(" %4u" % freq)
                        sys.stdout.write((" (%.2fs)      " % self.getWorkloadTime(workload, cpu))[:10])
            print ""

        print ""

    def printWorkloadTimeSummary(self, package):
        sys.stdout.write("Workload run times for package %s:\n" % package)
        sys.stdout.write("-------------------\n")
        workloads = self.workloads
        workloads.append("turbo")
        for workload in workloads:
            if workload in self.workloadTime:
                sys.stdout.write("%-13s" % workload)
                run = 1
                for run in range(1, self.getWorkloadTimesLength(workload)+1):
                    sys.stdout.write(" %7u " % run)
                    run += 1
                sys.stdout.write("\n")
                for cpu in self.packageToCpus[package]:
                    if self.getWorkloadTimes(workload, cpu):
                        sys.stdout.write("    cpu %3s: " % cpu)
                        for time in self.getWorkloadTimes(workload, cpu):
                            if time:
                                sys.stdout.write("%9.2f" % time)
                            else:
                                sys.stdout.write("    -    ")
                        print ""
                print ""
        print "\n"

    def adjustMinimumCheckTolerances(self):
        """ Adjusts minimum tolerances.

        Expectation is that self.cpuFlags has already been populated.
        """
        if self.hwpFlag in self.cpuFlags:
            self.minimumCheckLowerTolerance = self.lowerTolerance
            self.minimumCheckLowerTolerance = self.higherTolerance

    def enableAllCStates(self):
        print "Enabling all cstates"
        for cpu in range(0, self.numCPUs):
            for state in range(1, self.numCStates-1):
                Command("echo 0 > /sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state)).echo()

    def disableCState(self, cState):
        """disable c-state provided (and all higher) on all cpus
        """
        print "Disabling c-state %s and higher" % cState
        for cpu in range(0, self.numCPUs):
            for state in range(cState, self.numCStates-1):
                Command("echo 1 > /sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state)).echo()

    def outputCStateData(self, cState, filePath):
        Command("echo > %s" % filePath).echo()
        for cpu in range(0, self.numCPUs):
            timeVal = Command("cat /sys/devices/system/cpu/cpu%s/cpuidle/state%s/time" % (cpu, cState)).getString()
            Command('echo "%s %s" >> %s' % (cpu, timeVal, filePath)).echo()

    def runCstate(self):
        #Run the cstate test first for x86
        result = TestResult.WARN

        cpuArch = self.getCPUarch()
        if cpuArch and self.getCPUarch().find("x86") == 0:
            try:
                # get max number of cpus (zero based)
                self.numCPUs = int(Command("nproc").getString())

                # get max number of cstates (zero based)
                numCStates = 0
                cStatesCommand = Command("ls -1 /sys/devices/system/cpu/cpu0/cpuidle | wc -l")
                try:
                    numCStates = int(cStatesCommand.getString())
                except HwCertCommandException:
                    print("Warning - could not determine the number of cstates using command:\n%s" % cStatesCommand.command)

                try:
                    numCStates = int(numCStates)
                except ValueError:
                    print("Warning - could not convert %s to integer value" % numCStates)

                self.numCStates = numCStates

                # reset all cstates
                self.enableAllCStates()

                # iterate over c-states, disable cstate, wait a second to get values from all
                # cpus, wait another 30 seconds, get values again, and diff the results.
                # It is expected the values be consistent.  c-state zero cannot be disabled.
                # NOTE - if for some reason there are no c states detected, but process still gets to here,
                # self.numCStates should be 0 and the below for loop will not break, it just won't execute the inner block
                for x in range(1, self.numCStates-1):
                    self.disableCState(x)
                    time.sleep(1)
                    self.outputCStateData(x, "/tmp/start")
                    time.sleep(30)
                    self.outputCStateData(x, "/tmp/end")
                    if filecmp.cmp("/tmp/start", "/tmp/end"):
                        print "CPU c-state %s passed" % x
                        result = TestResult.PASS
                    else:
                        print "CPU c-state %s failed" % x
                        print "starting values:"
                        Command("cat /tmp/start").echo()
                        print "do not match ending values:"
                        Command("cat /tmp/end").echo()
                    self.enableAllCStates()

            except HwCertCommandException, exception:
                print("Warning: Cstate subtest did not pass, thus continuing testing further subtests.")

        return result


    def run(self):

        if self.runSubTest(self.runCstate, "Cstate") == TestResult.PASS:
            return True # If the Cstate test has passed do not execute the Pstate test

        #If the cstate test has failed run the pstate test

        if not self.runSubTest(self.getSystemCapabilities, "Capabilities"):
            return False

        # CERTPX-1425
        self.adjustMinimumCheckTolerances()

        # CERTPX-238
        if self.hwpFlag in self.cpuFlags:
            self.lowerTolerance = 5000 # Mhz

        result = True
        for package in self.packageToCpus.keys():

            print "Testing CPU Package " + package
            self.currentPackage = package

            self.waitForLull()

            if not self.runSubTest(self.runUserSpaceTests, "User Space, package %s" % package):
                result = False

            if not self.runSubTest(self.runOnDemandTests, "On Demand, package %s" % package):
                result =  False

            if not self.runSubTest(self.runPerformanceTests, "Performance, package %s" % package):
                result =  False

            self.restoreGovernors()

        if not self.runSubTest(self.runSummary, "Summary"):
            result =  False

        return result

class LoadProcess:
    def __init__(self, owner, cpu, aperfmperfSupport):
        self.owner = owner
        self.cpu = cpu
        self.workProcess  = None
        self.returnCode = None
        self.start_elapsed_time = None
        self.runTime = None
        self.effectiveFrequency = None
        self.aperfmperfSupport = aperfmperfSupport
        # signal.signal(signal.SIGCHLD, self.sigChldHandler)
        self.command = "./cpuscaling.py work "  # Use the Effective Frequency Interface and pi() calculation for a load process
        if self.aperfmperfSupport:
            self.command = "./aperf "               # Use aperf.c, which uses TSC/aperf/mperf and a square root calculation from libm

    def start(self):
        try:
            print "starting process for cpu %s" % self.cpu
            self.workProcess = Command("%s %s" % (self.command, self.cpu))
            print "using work process: %s" % self.workProcess.command
            self.workProcess.start()
            taskset = Command("taskset -pc %s %s" % (self.cpu, self.workProcess.getPID()))
            if self.owner.debug != Constants.off:
                taskset.echo()
            else:
                taskset.run()
        except HwCertCommandException, exception:
            print "Error: could not start load process"
            print exception
            print exception.command.printErrors()
            return False
        return True

    def getRunTime(self):
        return self.runTime

    def getEffectiveFrequency(self):
        return self.effectiveFrequency

    def sigChldHandler(self, signalNumber, stackFrame):
            print "Got signal %d" % signalNumber
            self.__stopTimer()
            self.owner.done(self.cpu)

    def isDone(self):
        if self.workProcess:
            self.returnCode = self.workProcess.poll()
            if self.returnCode is not None:
                while True:
                    line = self.workProcess.readline()
                    if line:
                        try:
                            if line.startswith("runTime: "):
                                self.runTime = float(line.split()[1])
                            elif line.startswith("freq: "):
                                self.effectiveFrequency = int(line.split()[1])
                        except (ValueError, IndexError):
                                pass # keep trying
                    else:
                        break
                return True
        # otherwise
        return False

    def getReturnCode(self):
        return self.returnCode


def pi():
        decimal.getcontext().prec = 500
        s = decimal.Decimal(1)
        h = decimal.Decimal(3).sqrt()/2
        n = 6
        for i in range(170):
            A = n*h*s/2  # A ... area of polygon
            s2 = ((1-h)**2+s**2/4)
            s = s2.sqrt()
            h = (1-s2/4).sqrt()
            n = 2*n
        # print i,":",A
        return True

if __name__ == "__main__":

    # this load process uses the Effective Frequency Interface (see aperfmperf.py)
    if "work" in sys.argv:
        effectiveFrequencyInterface = None
        effectiveFrequency = 0
        if len(sys.argv) > 2 and sys.argv[2].isdigit():
            cpuNumber = int(sys.argv[2])
            effectiveFrequencyInterface = EffectiveFrequency(cpuNumber)
            effectiveFrequencyInterface.start_sampling()
        (start_utime, start_stime, start_cutime, start_cstime, start_elapsed_time) = os.times()
        pi()
        (stop_utime, stop_stime, stop_cutime, stop_cstime, stop_elapsed_time) = os.times()
        print "" # make sure data is on its own line
        if effectiveFrequencyInterface:
            effectiveFrequency = effectiveFrequencyInterface.stop_sampling()
            if effectiveFrequency > 0:
                print "freq: %s" % effectiveFrequency
        runTime = stop_elapsed_time - start_elapsed_time
        print "runTime: %s" % (runTime, )
        sys.stdout.flush()
        sys.exit(0)

    test = CPUScalingTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
