#!/usr/bin/python
# Copyright (c) 2015 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Marek Kasik
#

import sys
import dbus.glib
import gobject as GObject
import glib as GLib

from rhcert.tags import Constants, DeviceClass, TestTag
from hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.redhatrelease import RedHatRelease

class FingerprintReaderTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "fingerprintreader")
        self.interactive = True
        self.deviceClass = DeviceClass.usb
        self.redHatRelease = RedHatRelease()


    def tags(self):
        return [TestTag.interactive, TestTag.certification]


    def getRequiredRPMs(self):
        return ['fprintd']


    def plan(self, planner):
        tests = list()
        try:
            bus = dbus.SystemBus()
            readers = bus.call_blocking('net.reactivated.Fprint',
                                         '/net/reactivated/Fprint/Manager',
                                         'net.reactivated.Fprint.Manager',
                                         'GetDevices',
                                         '',
                                         ())
        except dbus.exceptions.DBusException as e:
            return tests

        for reader in readers:
            self.reader = reader
            test = self.makeCopy()
            tests.append(test)

        return tests


    def quitVerification(self):
        self.proxy.Release('')
        self.proxy.DeleteEnrolledFingers('')
        self.verifySignal.remove()
        self.loop.quit()


    def retryVerification(self):
        self.unsuccessful_attempts += 1
        if self.unsuccessful_attempts < 3:
            self.proxy.VerifyStart('right-index-finger')
            if self.scan_type == 'swipe':
                self.statusMessage('  Swipe your right index finger through the \'' + self.name + '\' reader again please')
            else:
                self.statusMessage('  Place your right index finger on the reader \'' + self.name + '\' again please')
        else:
            self.quitVerification()


    def verifyStatusCb(self, result, done):
        print('  Verify result: ' + result)
        print('')
        try:
            self.proxy.VerifyStop()
            if done:
                if result == 'verify-match':
                    self.verifySuccess = True
                    self.quitVerification()

                elif result == 'verify-no-match':
                    self.retryVerification()

                else:
                    self.quitVerification()

            else:
                self.retryVerification()

        except dbus.exceptions.DBusException as e:
            print('Exception: %s' % e)
            self.quitVerification()


    def fingerprintReaderVerifyTestSource(self):
        try:
            self.unsuccessful_attempts = 0
            self.proxy.Claim('')
            self.verifySignal = self.proxy.connect_to_signal('VerifyStatus', self.verifyStatusCb)
            self.proxy.VerifyStart('right-index-finger')

            if self.scan_type == 'swipe':
                self.statusMessage('  Swipe your right index finger through the \'' + self.name + '\' reader to test verification please')
            else:
                self.statusMessage('  Place your right index finger on the reader \'' + self.name + '\' to test verification please')

        except dbus.exceptions.DBusException as e:
            print('Exception: %s' % e)
            self.loop.quit()

        return False


    def quitEnrollment(self):
        self.proxy.Release('')
        if not self.enrollSuccess:
            self.proxy.DeleteEnrolledFingers('')
        self.enrollSignal.remove()
        self.loop.quit()


    def retryEnrollment(self):
        self.unsuccessful_attempts += 1
        if self.unsuccessful_attempts < 3:
            self.proxy.EnrollStart('right-index-finger')
            if self.scan_type == 'swipe':
                self.statusMessage('  Swipe your right index finger through the \'' + self.name + '\' reader again please')
            else:
                self.statusMessage('  Place your right index finger on the reader \'' + self.name + '\' again please')
        else:
            self.quitEnrollment()


    def enrollStatusCb(self, result, done):
        print('  Enroll result: ' + result)
        print('')
        try:
            self.proxy.EnrollStop()
            if done:
                if result == 'enroll-completed':
                    self.enrollSuccess = True
                    self.quitEnrollment()

                elif result == 'enroll-failed':
                    self.retryEnrollment()

                else:
                    self.quitEnrollment()

            else:
                self.retryEnrollment()

        except dbus.exceptions.DBusException as e:
            print('Exception: %s' % e)
            self.quitEnrollment()


    def fingerprintReaderEnrollTestSource(self):
        try:
            self.unsuccessful_attempts = 0
            self.proxy.DeleteEnrolledFingers('')
            self.proxy.Claim('')
            self.enrollSignal = self.proxy.connect_to_signal('EnrollStatus', self.enrollStatusCb)
            self.proxy.EnrollStart('right-index-finger')

            if self.scan_type == 'swipe':
                self.statusMessage('  Swipe your right index finger through the \'' + self.name + '\' reader to test enrolling please')
            else:
                self.statusMessage('  Place your right index finger on the reader \'' + self.name + '\' to test enrolling please')

        except dbus.exceptions.DBusException as e:
            print('Exception: %s' % e)
            self.loop.quit()

        return False


    def runFingerprintReaderEnrollTest(self):
        self.enrollSuccess = False
        self.loop = GObject.MainLoop()
        GLib.idle_add(self.fingerprintReaderEnrollTestSource)
        self.loop.run()

        return self.enrollSuccess


    def runFingerprintReaderVerifyTest(self):
        self.verifySuccess = False
        self.loop = GObject.MainLoop()
        GLib.idle_add(self.fingerprintReaderVerifyTestSource)
        self.loop.run()

        return self.verifySuccess


    def run(self):
        success = True

        try:
            self.bus = dbus.SystemBus()
            self.name = self.bus.call_blocking('net.reactivated.Fprint',
                                                self.reader,
                                                'org.freedesktop.DBus.Properties',
                                                'Get',
                                                'ss',
                                                ('net.reactivated.Fprint.Device', 'name'))

            self.scan_type = self.bus.call_blocking('net.reactivated.Fprint',
                                                     self.reader,
                                                     'org.freedesktop.DBus.Properties',
                                                     'Get',
                                                     'ss',
                                                     ('net.reactivated.Fprint.Device', 'scan-type'))
            obj = self.bus.get_object('net.reactivated.Fprint', self.reader)
            self.proxy = dbus.Interface(obj, 'net.reactivated.Fprint.Device')
        except dbus.exceptions.DBusException as e:
            print('Exception: %s' % e)
            success = False

        if success:
            self.statusMessage('Testing \'' + self.name + '\' fingerprint reader:')
            success &= self.runFingerprintReaderEnrollTest()
            if success:
                success &= self.runFingerprintReaderVerifyTest()

        return success


if __name__ == '__main__':
    test = FingerprintReaderTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
