#!/usr/bin/python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys, os, time, fileinput

from rhcert.test import TestTag
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput
from rhcert.tags import DeviceClass

from hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.redhatrelease import RedHatRelease


class ProfilerTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "profiler")
        self.deviceClass = DeviceClass.processor
        self.forcedTimerConfig = False
        self.redHatRelease = RedHatRelease()

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        test = self.makeCopy()
        tests.append(test)
        return tests

    def getRequiredRPMs(self):
        return ["oprofile"]
    
    def logDevices(self, planner):
        devices = list()
        proc = ProcDeviceDetector()
        devices.extend(proc.getDevices(ProcDeviceDetector.procCPUInfo))
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices
        
    def initialize(self):
        self.getLinuxImage()
        self.resetNMIWatchdog()
        return True

    def old_run(self):


        try: 
            oprofile = Command("runtest --tool=oprofile")
            failureCount = oprofile.getInteger(regex="unexpected result[ \t]*(?P<count>\d+).*$", regexGroup="count", singleLine=False)
            print "%u failures were found" % failureCount
            return False
        except HwCertCommandException:
            # no failures were found
            pass

        oprofile.printOutput()

        return True

    def getLinuxImage(self):
        self.linuxVersion = Command("uname -r").getString()
        for path in ["/boot/vmlinux-%s", "/usr/lib/debug/lib/modules/%s/vmlinux"]:
            self.vmlinux = path % self.linuxVersion
            if os.path.exists(self.vmlinux):
                print "using linux image %s" % self.vmlinux
                break
            else:
                self.vmlinux = None
        
        if not self.vmlinux:
            print "Linux image for %s not found" % self.linuxVersion
            self.imageOption = "--no-vmlinux"
        else:    
            print "Using Linux image %s" % self.vmlinux
            self.imageOption = "--vmlinux=%s" % self.vmlinux
        sys.stdout.flush()
       
            
            
    def getOProfileVersion(self):
        command = Command("opcontrol -v")
        version = command.getString(regex="[^\d]*(?P<version>\d+\.\d+\.\d+[a-zA-Z_]*)", regexGroup="version")
        print "oprofile version: %s" % version
   
    def resetOProfile(self):
        print "\n==== START: Errors during reset may be ignored. ===="
        sys.stdout.flush()

        # start the daemon
        try:
            Command("opcontrol --shutdown").echo()
            Command("opcontrol --reset").echo()
        except HwCertCommandException, exception:
            print "Warning: "
            print exception
            pass
        
        try:
            command = Command("opreport")
            command.echo()
            # the above command should have error output
            print "Warning: opreport shouldn't report data after a reset"

        except HwCertCommandException, exception:
            # we expect an exception for error output
            print "Verified data has beed removed"
            sys.stdout.flush()
            sys.stderr.flush()
            pass
        
        try:
            Command("/bin/umount /dev/oprofile").echo()
            Command("/sbin/rmmod oprofile").echo()
            
            sys.stdout.flush()
            
            command = Command("lsmod")
            command.run()
            if "oprofile" in command.output:
                print "Warning: oprofile module not removed on system reset"
        except HwCertCommandException:
            pass
        
        print "^^^^ END: Errors during reset may be ignored. ^^^^\n"
        sys.stdout.flush()

       

        return True
    
    def startOProfile(self):
        print "oprofile start: initializing..."
        Command("opcontrol --init").echo()
        tries = 5
        command = Command("lsmod")
        command.getString(regex="oprofile", singleLine=False)
        
        print "checking /dev/oprofile filesystem presence"
        if os.path.exists("/dev/oprofile/cpu_type"):
            print "oprofile filesystem mounted"
        else:
            print "oprofile filesystem not mounted"
            return False
        
        command = Command("opcontrol --setup %s" % self.imageOption)
        command.echo()
        command = Command("opcontrol --start")
        command.echo()
        self.getOProfileVersion()
        return True
    
    def startDaemon(self):
        # this starts daemon, but doesn't start profiling
        print "starting opcontrold"
        
        # os.system is used here to work around a hanging bug on ia64
        # we don't care if it the call reports failure, so return value is ignored
        os.system("opcontrol --start-daemon %s" % self.imageOption)
        time.sleep(5)
        command = Command("ps ax")
        try:
            command.getString(regex="oprofiled", singleLine=False)
        except:
            print "Error: oprofiled did not start"
            sys.stdout.flush()
            return False
        print "started"
        sys.stdout.flush()
        return True
        
    def report(self):
        try:
            Command("opcontrol --dump").echo()
            try:
                Command("opreport").echo()
            except HwCertCommandErrorOutput, exception:
                print "Warning: %s" % exception
                exception.command.printErrors()
        except HwCertCommandException, exception:
            print exception
            return False
        return True
    
    def findTimerConf(self):
        for self.timerConfPath in ["/etc/modprobe.conf", "/etc/modprobe.d/oprofile.conf"]:
            if os.path.exists(self.timerConfPath):
                break

        if not self.timerConfPath and os.path.exists("/etc/modprobe.d"):
            self.timerConfPath = "/etc/modprobe.d/oprofile.conf"
            try:
                timerConf = open(self.timerConfPath, "w")
                self.timerConfPath.close()
            except IOError, e:
                print "Error: could not create %s" % self.timerConfPath
                print e
                self.timerConfPath = None

    def checkTimerConf(self):
        for self.timerConfPath in ["/etc/modprobe.conf", "/etc/modprobe.d/oprofile.conf"]:
            if os.path.exists(self.timerConfPath):
                modprobeConf = open(self.timerConfPath, "r")
                for line in modprobeConf.readlines():
                    if line.startswith("options oprofile"):
                        return True
                # found file, but not the setting
                return False
        # otherwise
        return False

    def forceTimerConf(self):
        if not self.checkTimerConf():
            print "Warning: Forcing oprofile module timer option"
            try:
                modprobeConf = open(self.timerConfPath, "a")
                modprobeConf.write("options oprofile timer=1\n")
                modprobeConf.close()
                self.forcedTimerConfig = True
            except:
                return False

        return True

    def removeTimerConf(self):
        if self.checkTimerConf():
            for line in fileinput.input(self.timerConfPath, "inplace=1"):
                if not line == "options oprofile timer=1\n":
                    print line.rstrip("\n")
                    
    def resetNMIWatchdog(self):
        try:
            self.nmiWatchdog = Command("cat /proc/sys/kernel/nmi_watchdog").getString()
            print "NMI Watchdog = %s" % self.nmiWatchdog
            if self.nmiWatchdog != "0":
                print "reseting NMI watchdog"
                Command("echo 0 > /proc/sys/kernel/nmi_watchdog").echo()
            return True
        except HwCertCommandException, exception:
            print "Error: could not read/set NMI Watchdog"
            print exception
        
        return False
            
    def restoreNMIWatchdog(self):
        try:
            nmiWatchdog = Command("cat /proc/sys/kernel/nmi_watchdog").getString()
            if self.nmiWatchdog != "0" and nmiWatchdog != self.nmiWatchdog:
                print "restoring NMI watchdog"
                Command("echo %s > /proc/sys/kernel/nmi_watchdog" % self.nmiWatchdog).echo()
            return True
        except HwCertCommandException, exception:
            print "Error: could not restore NMI Watchdog"
            print exception
        
        return False
    
    def runOProfileStartup(self):
        if not self.runSubTest(self.resetOProfile, "Reset"):
            return False
        if not self.runSubTest(self.startDaemon, "Start Daemon"):
            return False
        if not self.runSubTest(self.startOProfile, "Start OProfile"):
            return False
        
        return True


    def oPerfTest(self):
        self.getLinuxImage()
        vmlinux_option = ""
        if self.vmlinux:
            vmlinux_option = "-k %s" % self.vmlinux
        command = Command("timeout -s INT 5s operf %s -l -s" % vmlinux_option)
        #operf sends the text 'operf: starting' and other non-errors to stderr
        try:
            command.echo()
        except HwCertCommandErrorOutput, exception:
            print exception
            exception.command.printErrors()
        #command.echoIgnoreErrors()
        #opreport sends a message about where the files are being read from to stderr
        try:
            Command("opreport").echo()
        except HwCertCommandErrorOutput, exception:
            print exception
            exception.command.printErrors()
        return True


    def run(self):

        success = True
        if "ARM" in self.redHatRelease.product:
            return self.runSubTest(self.oPerfTest, "Running OProfile Test")

        if not self.runSubTest(self.initialize, "Initialize"):
            success = False

        if success and not self.runOProfileStartup():
            success = False

        # if report fails, re-try with a forced timer config
        if not self.runSubTest(self.report, "Report"):
            print "Re-try testing with forced timer configuration:"
            if self.forceTimerConf():
                success = True
                if not self.runOProfileStartup():
                    success = False
                else:
                    if not self.runSubTest(self.report, "Report"):
                        success = False
                self.removeTimerConf()
            else:
                print "could not test with forced timer configuration."


        self.restoreNMIWatchdog()

        # reset - ignore if it fails
        try:
            self.runSubTest(self.resetOProfile, "Reset")
        except:
            pass    

        return success

if __name__ == "__main__":
    test = ProfilerTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
   
   


