#!/usr/bin/python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, sys, time, syslog, re

from rhcert.tags import Constants, DeviceClass, TestTag
from rhcert.controller import Controller

from hwcert.hwcert.hardwareTest import HardwareTest

class SuspendTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "suspend")
        self.acpiListen = None
        self.hwcertController = Controller()
        self.interactive = True
        self.deviceClass = DeviceClass.battery
        self.priority = 1 # run suspend before any other test
        self.mandatory = Constants.pre
        #  451154 - failures on unknown suspend methods
        self.knownMethods = [Constants.memory, Constants.disk]
        self.suspendMessage = dict()
        self.suspendMessage[Constants.memory] = "Suspending console"
        self.suspendMessage[Constants.disk] = "Creating hibernation image"

    def tags(self):
        return [TestTag.interactive, TestTag.certification, TestTag.portable]

    def plan(self, planner):
        tests = list()
        acAdapterDevice = None
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            if "power_supply" in device.getProperty("SUBSYSTEM") and "BAT" in device.getProperty("POWER_SUPPLY_NAME"):
                test = self.makeCopy()
                test.setDevice(device)
                tests.append(test)
                # 450297: suspend test planned once per battery
                # change to only plan this test once, no matter how many
                # batteries are reported
                break
        return tests

    def getRequiredRPMs(self):
        return ["pm-utils"]

    def getSuspendMethods(self):
        sysPowerState = open("/sys/power/state", "r")
        if sysPowerState:
            while True:
                line = sysPowerState.readline()
                if not line:
                    break
                states = line.split()
                #  451154 - failures on unknown suspend methods
                methods = list()
                for state in states:
                    if state in self.knownMethods:
                        methods.append(state)
                if len(methods)>0:
                    return methods
        # otherwise
        print "Error: could not determine allowable suspend methods"
        return None

    def suspendWithOSCommand(self, method):
        """ suspend via pm-suspend/pm-hibernate """
        if method == Constants.memory:
            if os.system("pm-suspend") != 0:
                return False
        elif method == Constants.disk:
            if os.system("pm-hibernate") != 0:
                return False
        else:
            print "Error: unknown suspend method %s" % method
            return False

        # otherwise
        return True

    def suspendWithOSCommand2(self, method):
        """ suspend using a write operation to /sys/power/state """
     # if not os.system("echo \"%s\" > /sys/power/state" % method) is 0:
        sysPowerState = open("/sys/power/state", "w")
        if sysPowerState:

            sysPowerState.write(method)
            sysPowerState.close()
        else:
            print "Error: could not suspend the system"
            return False
        return True

    def suspend(self, source, method):
        print "This test will suspend the operating system."
        print "Please resume by pressing the power button after suspend has completed."
        sys.stdout.flush()
         # start acpi_listen
        # if self.acpiListen:
        #    self.acpiListen.close()
        # self.acpiListen = os.popen("acpi_listen")
        markerName = "%s-%s-%s" % (self.getPath(), source, method)
        syslog.syslog(self.hwcertController.getSystemLogMarker(markerName, "begin"))
        if source == Constants.OSCommand:
            if not self.promptConfirm(" suspend? "):
                return False
            waitTime = 5
            print "Suspending in %d sec" % waitTime
            sys.stdout.flush()
            time.sleep(waitTime)
            return self.suspendWithOSCommand(method)

        # otherwise
        if source == Constants.FunctionKey:
            if not self.promptConfirm("Does this system have a function key (Fn) to suspend the system to %s?" % method):
                print "Warning: suspend test to %s not run from function key" % method
                return False
            if not self.promptConfirm("Are you ready to press the function key (Fn) to suspend the system to %s. Answer the question and then press the function key (Fn) if you want to suspend the system" % method):
                return False
            return True

        # otherwise
        print "Error: unknown suspend source"
        return False

    def checkResume(self, source, method):
        self.verifiedResumeMethod = None
        sys.stdout.flush()
        self.promptConfirm("Has resume completed? ")
        markerName = "%s-%s-%s" % (self.getPath(), source, method)
        syslog.syslog(self.hwcertController.getSystemLogMarker(markerName, "end"))

        #check system log for acpi messages
        log = self.hwcertController.getSystemLog("%s-%s-%s" % (self.getPath(), source, method))

        if source == Constants.FunctionKey:
            # check Function Key message (given by systemd) first:
            try:
                suspendMessage = self.suspendMessage[method]
            except KeyError:
                print "Error: unknown suspend method %s" % method
                return False

            if not suspendMessage in log:
                print "Error: could not verify suspend"
                return False

        # check suspend/hibernate message (given by kernel)
        suspendMessage = "Disabling non-boot CPUs"
        if not suspendMessage in log:
            print "Error: could not verify suspend"
            return False

        # otherwise
        print "Verified suspend"

        # verifying resuming
        resumeMessage = "Enabling non-boot CPUs"
        if not resumeMessage in log:
            print "Error: could not verify resume"
            return False

        # check that they used the correct method
        self.verifiedResumeMethod = Constants.memory
        methodMessage = re.compile("(swsusp|PM): Need to copy")
        for line in log.split('\n'):
            if methodMessage.search(line):
                self.verifiedResumeMethod = Constants.disk
                break
        if method != self.verifiedResumeMethod:
            print "Error: The suspend method was %s and should have been %s" % (self.verifiedResumeMethod, method)
            return False

        # otherwise
        print "Verified resume from %s" % self.verifiedResumeMethod
        return True

    def runSuspendResume(self, source, method):
        if self.suspend(source, method):
            return self.checkResume(source, method)
        elif source == Constants.FunctionKey:
            # allow user to opt-out of fn key tests, since system may not have them
            self.verifiedResumeMethod = method
            return True
        #otherwise
        return False


    def run(self):

        if self.getMode() == Constants.auto:
            print "Error: suspend test can not be run in auto mode."
            return False

        methods = self.getSuspendMethods()
        if not methods:
            return False

        # forced mode - being run as preparations for other tests
        if self.getMode() == Constants.forced:
            result = True
            if not self.runSuspendResume(Constants.OSCommand, methods[0]):
                result = False
            elif self.verifiedResumeMethod != methods[0]:
                result = False
            return result

        # otherwise, full test
        result = True
        for source in [Constants.OSCommand, Constants.FunctionKey]:
            for method in methods:
                self.markOutput(name=source + "-" + method)
                if not self.runSuspendResume(source, method):
                    result = False
                    self.markSummary(Constants.FAIL)
                elif self.verifiedResumeMethod != method:
                    result = False
                    self.markSummary(Constants.FAIL)
                else:
                    self.markSummary(Constants.PASS)

        self.closeOutput()
        return result


if __name__ == "__main__":
    test = SuspendTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
