# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.#
# Author: Irina Boverman
#

import os, re, sys, string, commands, shutil, platform

from rhcert.command import Command, HwCertCommandException
from rhcert.tags import DeviceClass, TestTag
from rhcert.redhatrelease import RedHatRelease

from hwcert.hwcert.hardwareTest import HardwareTest

class VideoTest(HardwareTest):
# A wrapper for x11perf; while we're not actually looking for benchmark numbers
# we may perhaps one day start to compare and contrast against them to ensure
# there are no performance regressions.  Until then; we're using x11perf's
# use of X primitives to check for proper operation.

    def __init__(self):
        HardwareTest.__init__(self, "video")

        self.Xconfig_tool = ""
        self.Xconfig_flag = ""
        self.depth = 0
        self.interface = "video-card"
        self.Xconfigfile = "/tmp/hwcertXconfig"
        self.tmpFile = "/tmp/hwcert.x11perf"
        self.deviceClass = DeviceClass.display
        self.redHatRelease = RedHatRelease()

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        properties = dict()
        properties["PCI_CLASS"] = "30000"
        tests = self.planWith(planner.getUdevDeviceDetector(), properties, allDevices=False, strict=True)
        return tests

    def getRequiredRPMs(self):
        rpms = ["xorg-x11-apps"]
        rpms.append("xorg-x11-utils") # for xdpyinfo and xvinfo
        rpms.append("glx-utils") # for xdriinfo and glxinfo
        rpms.append("xorg-x11-server-Xorg") # for X
        rpms.append("xorg-x11-xinit") # for startx

        return rpms

    def settool(self):
        self.Xconfig_tool = None
        return True

    def getfile(self):
        if self.Xconfig_tool:
            status = os.system(self.Xconfig_tool + " -o " + self.Xconfigfile + "  --noui > /dev/null")
            if status != 0:
                print "Error: could not obtain config file"
                return False
        else: # check the log
            self.Xconfigfile = None
            logPath = "/var/log/Xorg.0.log"
            if os.path.exists(logPath):
                log = open(logPath)
                pattern = re.compile("\(==\) Using config file: \"(?P<filepath>[^\"]+)\"")
                for line in log.readlines():
                    match = pattern.search(line)
                    if match:
                        self.Xconfigfile = match.group("filepath")
                        print "Using config file " + self.Xconfigfile
                        break
                log.close()
            if not self.Xconfigfile:
                print "No config file found - using Default Config"
        return True

    def setdepth(self):
        if self.Xconfigfile:
            (status, self.depth) = commands.getstatusoutput("grep DefaultDepth " + self.Xconfigfile +
                                                        " | awk ' { print $2; } '")
            if status != 0:
                print "Error: could not obtain default depth from config file"
                return False
            try:
                self.depth = int(self.depth)
                print "Depth set to %u" % self.depth
            except:
                print "Error: could not convert depth %s to integer" % self.depth
                self.depth = None
                return False
        return True

    def setflag(self):
        # Adapt for flag changes
        (status, self.Xconfig_flag) = commands.getstatusoutput("X --help 2>&1 | grep xf86config -q && " +
                                                               "echo -xf86config || echo -config")
        if status != 0:
            print "FAILED to obtain config flag"
            return False
        print "Depth flag: " + self.Xconfig_flag
        return True

    def startx(self):
        try: os.remove(self.tmpFile)
        except: pass
        startxCommand = "startx ./x11perftest -- :2 "
        if self.Xconfigfile:
            startxCommand += "%s %s " % (self.Xconfig_flag, self.Xconfigfile)
        if self.depth:
            startxCommand +=  "-depth %u " % self.depth

        try:
            print "Running: " + startxCommand
            startx = Command(startxCommand + " &> /dev/null")
            startx.echo()
        except HwCertCommandException, e:
            print "Error: startx failed;"
            print e
            print startx.printErrors()
            return False

        # copy the x11perf log into the output file
        if os.path.exists("/var/log/Xorg.0.log"):
            shutil.copy("/var/log/Xorg.0.log", ".")
        if os.path.exists("/var/log/Xorg.2.log"):
            shutil.copy("/var/log/Xorg.2.log", ".")
        for filename in os.listdir("."):
            if "Xorg" in filename:
                xlog = open(filename)
                while 1:
                    line = xlog.readline()
                    if line:
                        print line
                    else:
                        break
                xlog.close();

        return True

    def xstatus(self):
        try:
            resultFile = open(self.tmpFile)
            result = 1 # error
            while 1:
                line = resultFile.readline()
                if line:
                    result = string.atoi(line)
                else:
                    break
            resultFile.close()
            if result != 0:
                sys.stdout.write("Error: x11perf returned %d" % result)
                return False
        except IOError:
            print "Error: x11perf did not run"
            return False
        return True

    def checkDepthAndResolution(self):
        success = True
        # try to get the current screen resolution and color depth first
        try:
            # inclue display flag to ensure it is set
            xdpyinfo = Command("xdpyinfo -display :0")
            xdpyinfo.run()
            for line in xdpyinfo.output:
                if "dimensions:" in line:
                    screenResolution = line.split(" ")[6].split("x")
                    horizontalResolution = screenResolution[0]
                    verticalResolution = screenResolution[1]
                if "depth of root window:" in line:
                    colorDepth = line.split(" ")[9]
            print "\nScreen resolution: %sx%s" % (horizontalResolution,verticalResolution)
            print   "      Color depth: %s" % colorDepth
            # minimum requirement is 1024x768@24bit, but on some wide-screen
            # systems, the vertical resolution maybe smaller than 768
            if int(horizontalResolution) < 1024:
                print "\nWarning ! Current screen resolution: %sx%s does not meet the minimum requirements !\n" % (horizontalResolution,verticalResolution)
                success = False
            if int(colorDepth) < 24:
                print "\nWarning ! Current color depth: %s does not meet the minimum requirements !\n" % colorDepth
                success = False
        except:
            # record error and continue
            print "\nWarning ! FAILED to obtain current screen resolution and color depth !"
            success = False

        return success

    def checkXModules(self):
        logPath = "/var/log/Xorg.0.log"
        log = open(logPath, "r")
        if not log:
            print "Error: could not open %s" % logPath
            return False

        success = True
        print "Checking loaded X modules -----------------------\n"
        print "X log file: %s" % logPath
        for line in log.readlines():
            # this presumes Loading lines that include a slash means a path
            if line.find("(II) Loading /") >= 0:
                # is there a oneline way to do this?
                xmoduleFile = line.split()
                xmoduleFile = xmoduleFile[-1]

                try:
                    Command("rpm -Vf %s" % xmoduleFile).run()
                    self.checkXModuleVendorAndBuildHost(xmoduleFile)
                except Exception, e:
                    print "Error: RPM verification failed for X module %s" % xmoduleFile
                    print e
                    success = False

        print "-------------------------------------------------\n"
        log.close()
        return success

    def checkXModuleVendorAndBuildHost(self, xmoduleFile):
        # Maybe this should be in a library; this function lifted from info.py
        result = True
        goodVendorList = ["Red Hat", "Red Hat, Inc."]
        warnVendorList = ["Fedora", "Fedora Project"]
        try:

            vendor = Command("rpm -qf %s --qf %%{VENDOR}" % xmoduleFile).getString()
            if vendor in goodVendorList:
                pass
            elif vendor in warnVendorList:
                print "Warning: %s module %s found" % (vendor, xmoduleFile)
            else:
                print "Error: Non-Red Hat vendor %s for module %s" % (vendor, xmoduleFile)
                result = False

            buildhost = Command("rpm -qf %s --qf %%{BUILDHOST}" % xmoduleFile).getString()
            if "redhat.com" not in buildhost:
                print "Error: X module %s was built on %s and not built at Red Hat." % (xmoduleFile, buildhost)
                result = False
        except HwCertCommandException, exception:
                print "Error: could not determine X module packager for  %s" % xmoduleFile
                print exception
                result = False

        return result

    def logDriverInfo(self):
        libDir = "/usr/lib64" if platform.architecture()[0] == "64bit" else "/usr/lib"
        logPath = "/var/log/Xorg.0.log"
        log = open(logPath, "r")
        if not log:
            print "Error: could not open %s" % logPath
            return
        print "\nDriver Info from %s:\n-------------------------------------------------" % logPath
        copying = False
        entryCount = 0
        while 1:
            line = log.readline()
            if line:
                if not copying:
                    if line.find("%s/xorg/modules/drivers" % libDir) >= 0:
                        copying = True
                # stop on second "(II)" or blank line
                else:
                    if line.find("(II)") >= 0:
                        entryCount += 1
                    if line.strip() == "" or entryCount > 1:
                        copying = False
                        entryCount = 0
                        print ""
                if copying:
                    sys.stdout.write(line)
            else:
                log.close()
                break
        print "-------------------------------------------------"

        # Should this output be stored in attachments?
        Command("xvinfo").echoIgnoreErrors()
        print "-------------------------------------------------"
        Command("xdriinfo").echoIgnoreErrors()
        print "-------------------------------------------------"
        Command("glxinfo").echoIgnoreErrors()
        print "-------------------------------------------------"

    def checkConnections(self):
        sys.stdout.write("Checking for displays...")
        xrandr = Command("xrandr")
        try:
            xrandr.echo()
            sys.stdout.flush()
            # remove the check, as it  is unrelable
            # xrandr.getString(" connected", singleLine=False)
            # print "   done."
            # sys.stdout.flush()
        except HwCertCommandException, e:
            # print "\nError: no displays connected"
            print e
            # return FAILED

        return True

    def run(self):
        print "Starting video test"
        print "on device udi: %s" % self.getUDI()
        result = False
        if (self.checkConnections() and self.settool() and self.getfile() and
                self.setdepth() and self.setflag() and self.startx() and self.xstatus() and
                self.checkDepthAndResolution() and self.checkXModules()):
            self.logDriverInfo()
            result = True
            print "Video test on device %s passed." % (self.interface)
        else:
            print "Video test on device %s FAILED." % (self.interface)
        return result

