# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import DeviceClass, TestTag
from rhcert.command import Command, HwCertCommandException

from hwcert.hwcert.networkTest import NetworkTest
from hwcert.hwcert.iwconfig import IWConfig


class WLanTest(NetworkTest):

    def __init__(self, path=None):
        if path:
            path = "wlan/" + path
        else:
            path = "wlan"
        NetworkTest.__init__(self, path)
        self.requiredSupport = None
        self.procNetWireless = "/proc/net/wireless"
        self.deviceClass = DeviceClass.network_interface

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.network, TestTag.wlan]

    def plan(self, planner):
        # get the wireless device name
        iwconfig = IWConfig()
        interfaces = iwconfig.getInterfaces()
        tests = dict()
        testList = list()
        if interfaces:
            devices = self.getNetworkDevices(planner)
            for (logicalDevice, device) in devices.items():
                if logicalDevice in interfaces and self.isBestType(iwconfig.getType()):
                    test = self.makeCopy()
                    test.setDevice(device)
                    test.setLogicalDeviceName(logicalDevice)
                    tests[logicalDevice] = test
            testList = list()
            testList.extend(tests.itervalues())
        return testList

    def isBestType(self, supportedTypes):
        """ return True if this type is the best in the list """
        if supportedTypes:
            for type in ["ac","n", "g"]:
                if type in supportedTypes:
                    return type == self.requiredSupport
        # if ac, n or g is not supported and we don't require support, we're the best
        return not self.requiredSupport

    def getRequiredRPMs(self):
        """Returns list of strings representing required packages.
        wireless-tools for iw command
        rfkill for rfkill command
        """
        rpms = [ "wireless-tools", "rfkill" ]
        return rpms

    def verify(self):
        try:
            log = Command("iwconfig %s" % self.getLogicalDeviceName())
            log.run()
            return True
        except HwCertCommandException, exception:
            print "Error: get configuration for %s" % self.procNetWireless
            print exception
            exception.command.printErrors()

        return False

    def bringUpDevice(self):
        try:
            ifup = Command("ifup %s"  % self.getLogicalDeviceName())
            ifup.echo()
        except HwCertCommandException, exception:
            print "Warning: could not bring up %s" % self.getLogicalDeviceName()
            print exception
            return False

        return True


    # 1. output of /proc/net/wireless should provide wireless interfaces to expect
    # and also their stats in procfs
    def logProcNetWireless(self):
        try:
            log = Command("cat " + self.procNetWireless)
            log.echo()
            return True
        except HwCertCommandException, exception:
            print "Error: could not log %s" % self.procNetWireless
            print exception
            exception.command.printErrors()

        return False

    # 2. following command should provide list of access points in scan range of
    # <dev> and properties of AP
    def scan(self):
        commands = ["iwconfig %s" % self.getLogicalDeviceName(), "iwlist %s scan" % self.getLogicalDeviceName()]
        result = True
        for command in commands:
            try:
                scan = Command(command)
                scan.echo()
            except   HwCertCommandException, exception:
                print "Error: %s:" % command
                print exception
                exception.command.printErrors()
                result = False

        return result

    # 3. Following iwgetid commands are useful for simple outputs
    def iwGetID(self):
        commands = ["iwgetid %s" % self.getLogicalDeviceName(), "iwgetid --protocol %s"  % self.getLogicalDeviceName(), "iwgetid --mode %s" % self.getLogicalDeviceName() ]
        result = True
        for command in commands:
            try:
                iwgetid = Command(command)
                iwgetid.echo()
            except   HwCertCommandException, exception:
                print "Error: %s:" % command
                print exception
                exception.command.printErrors()
                result = False
        return result

    def checkSupport(self):
        iwconfig = IWConfig()
        if self.requiredSupport in iwconfig.getType():
            print "Wireless %s is supported." % self.requiredSupport
            return True
        # otherwise
        print "Error: wireless %s is not supported by %s" % (self.requiredSupport, iwconfig.getType())
        return False

    def run(self):

        result = True
        if not self.runSubTest(self.logProcNetWireless, "Log proc", "log " + self.procNetWireless):
            result = False

        if not self.runSubTest(self.scan, "Scan Wireless Networks"):
            result = False

        if not self.runSubTest(self.iwGetID, "Get wireless connection information"):
            result = False

        if self.requiredSupport and not self.runSubTest(self.checkSupport, "Check Support"):
            result = False

        if NetworkTest.run(self) != True:
            result = False

        return result

class WirelessGTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessG")
        self.requiredSupport = "g"
        self.enforceSupport = True
        self.interfaceSpeed = 22 # speed in Mb/sec
        self.enforceSpeed = True


class WirelessNTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessN")
        self.requiredSupport = "n"
        self.enforceSupport = True
        self.interfaceSpeed = 100 # speed in Mb/sec
        self.enforceSpeed = True

class WirelessACTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessAC")
        self.requiredSupport = "ac"
        self.enforceSupport = True
        self.interfaceSpeed = 300 # speed in Mb/sec
        self.enforceSpeed = True


